# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" 'audio.spectrogram' module within the ketos library.

    This module provides utilities to work with spectrograms.

    Spectrograms are two-dimensional visual representations of 
    sound waves, in which time is shown along the horizontal 
    axis, frequency along the vertical axis, and color is used 
    to indicate the sound amplitude. Read more on Wikipedia:
    https://en.wikipedia.org/wiki/Spectrogram

    The module contains the parent class Spectrogram, and four
    child classes (MagSpectrogram, PowerSpectrogram, MelSpectrogram, 
    CQTSpectrogram), which inherit methods and attributes from the 
    parent class.

    Note, however, that not all methods (e.g. crop) work for all 
    child classes. See the documentation of the individual methods 
    for further details.

    Contents:
        Spectrogram class:
        MagSpectrogram class:
        PowerSpectrogram class:
        MelSpectrogram class:
        CQTSpectrogram class
"""
import os
import copy
import librosa
import numpy as np
from scipy.signal import get_window
from scipy.fftpack import dct
from scipy import ndimage
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from ketos.audio.waveform import Waveform
import ketos.audio.utils.misc as aum
from ketos.audio.utils.axis import LinearAxis, Log2Axis
from ketos.audio.annotation import AnnotationHandler
from ketos.audio.utils.filter import enhance_signal, reduce_tonal_noise
from ketos.audio.base_audio import BaseAudio, segment_data


def add_specs(a, b, offset=0, scale=1, make_copy=False):
    """ Place two spectrograms on top of one another by adding their 
        pixel values.

        The spectrograms must be of the same type, and share the same 
        time resolution. 
        
        The spectrograms must have consistent frequency axes. 
        For linear frequency axes, this implies having the same 
        resolution; for logarithmic axes with base 2, this implies having 
        the same number of bins per octave minimum values that differ by 
        a factor of :math:`2^{n/m}` where :math:`m` is the number of bins 
        per octave and :math:`n` is any integer. No check is made for the 
        consistency of the frequency axes.

        Note that the attributes filename, offset, and label of spectrogram 
        `b` is being added are lost.

        The sum spectrogram has the same dimensions (time x frequency) as 
        spectrogram `a`.

        Args:
            a: Spectrogram
                Spectrogram
            b: Spectrogram
                Spectrogram to be added
            offset: float
                Shift spectrogram `b` by this many seconds relative to spectrogram `a`.
            scale: float
                Scaling factor applied to signal that is added
            make_copy: bool
                Make copies of both spectrograms, leaving the orignal instances 
                unchanged by the addition operation.

        Returns:
            ab: Spectrogram
                Sum spectrogram
    """
    assert a.type == b.type, "It is not possible to add spectrograms with different types"
    assert a.time_res() == b.time_res(), 'It is not possible to add spectrograms with different time resolutions'

    # make copy
    if make_copy:
        ab = a.deepcopy()
    else:
        ab = a

    # compute cropping boundaries for time axis
    end = a.duration() - offset

    # determine position of b within a
    pos_x = a.time_ax.bin(offset, truncate=True) #lower left corner time bin
    pos_y = a.freq_ax.bin(b.freq_min(), truncate=True) #lower left corner frequency bin

    # crop spectrogram b
    b = b.crop(start=-offset, end=end, freq_min=a.freq_min(), freq_max=a.freq_max(), make_copy=make_copy)

    # add the two images
    bins_x = b.data.shape[0]
    bins_y = b.data.shape[1]
    ab.data[pos_x:pos_x+bins_x, pos_y:pos_y+bins_y] += scale * b.data

    return ab

def mag2pow(img, num_fft):
    """ Convert a Magnitude spectrogram to a Power spectrogram.

        Args:
            img: numpy.array
                Magnitude spectrogram image.
            num_fft: int
                Number of points used for the FFT.
        
        Returns:
            : numpy.array
                Power spectrogram image
    """
    return (1.0 / num_fft) * (img ** 2)

def mag2mel(img, num_fft, rate, num_filters, num_ceps, cep_lifter):
    """ Convert a Magnitude spectrogram to a Mel spectrogram.

        Args:
            img: numpy.array
                Magnitude spectrogram image.
            num_fft: int
                Number of points used for the FFT.
            rate: float
                Sampling rate in Hz.
            num_filters: int
                The number of filters in the filter bank.
            num_ceps: int
                The number of Mel-frequency cepstrums.
            cep_lifters: int
                The number of cepstum filters.
        
        Returns:
            mel_spec: numpy.array
                Mel spectrogram image
            filter_banks: numpy.array
                Filter banks
    """
    power_spec = mag2pow(img, num_fft)
    
    low_freq_mel = 0
    high_freq_mel = (2595 * np.log10(1 + (rate / 2) / 700))  # Convert Hz to Mel
    mel_points = np.linspace(low_freq_mel, high_freq_mel, num_filters + 2)  # Equally spaced in Mel scale
    hz_points = (700 * (10**(mel_points / 2595) - 1))  # Convert Mel to Hz
    bin = np.floor((num_fft + 1) * hz_points / rate)

    fbank = np.zeros((num_filters, int(np.floor(num_fft / 2 + 1))))
    for m in range(1, num_filters + 1):
        f_m_minus = int(bin[m - 1])   # left
        f_m = int(bin[m])             # center
        f_m_plus = int(bin[m + 1])    # right

        for k in range(f_m_minus, f_m):
            fbank[m - 1, k] = (k - bin[m - 1]) / (bin[m] - bin[m - 1])
        for k in range(f_m, f_m_plus):
            fbank[m - 1, k] = (bin[m + 1] - k) / (bin[m + 1] - bin[m])

    filter_banks = np.dot(power_spec, fbank.T)
    filter_banks = np.where(filter_banks == 0, np.finfo(float).eps, filter_banks)  # Numerical Stability
    filter_banks = 20 * np.log10(filter_banks)  # dB
    
    mel_spec = dct(filter_banks, type=2, axis=1, norm='ortho')[:, 1 : (num_ceps + 1)] # Keep 2-13

    (nframes, ncoeff) = mel_spec.shape
    n = np.arange(ncoeff)
    lift = 1 + (cep_lifter / 2) * np.sin(np.pi * n / cep_lifter)
    mel_spec *= lift  
    
    return mel_spec, filter_banks

def load_audio_for_spec(path, channel, rate, window, step,\
            offset, duration, resample_method, id=None):
    """ Load audio data from a wav file for the specific purpose of computing 
        the spectrogram.

        The loaded audio covers a time interval that extends slightly beyond 
        that specified, [offset, offset+duration], as needed to compute the 
        full spectrogram without zero padding at either end. If the lower/upper 
        boundary of the time interval coincidences with the start/end of the 
        audio file so that no more data is available, we pad with zeros to achieve
        the desired length. 

        Args:
            path: str
                Path to wav file
            channel: int
                Channel to read from. Only relevant for stereo recordings
            rate: float
                Desired sampling rate in Hz. If None, the original sampling rate will be used
            window: float
                Window size in seconds that will be used for computing the spectrogram
            step: float
                Step size in seconds that will be used for computing the spectrogram
            offset: float
                Start time of spectrogram in seconds, relative the start of the wav file.
            duration: float
                Length of spectrogrma in seconds.
            resample_method: str
                Resampling method. Only relevant if `rate` is specified. Options are
                    * kaiser_best
                    * kaiser_fast
                    * scipy (default)
                    * polyphase
                See https://librosa.github.io/librosa/generated/librosa.core.resample.html 
                for details on the individual methods.
            id: str
                Unique identifier (optional). If None, the filename will be used.

        Returns:
            audio: Waveform
                The audio signal
            seg_args: tuple(int,int,int,int)
                Input arguments for :func:`audio.utils.misc.segment`
    """
    if rate is None:
        rate = librosa.get_samplerate(path) #if not specified, use original sampling rate

    file_duration = librosa.get_duration(filename=path) #get file duration
    file_len = int(file_duration * rate) #file length (number of samples)
    
    assert offset < file_duration, 'Offset exceeds file duration'
    
    if duration is None:
        duration = file_duration - offset #if not specified, use file duration minus offset

    # compute segmentation parameters
    seg_args = aum.segment_args(rate=rate, duration=duration,\
        offset=offset, window=window, step=step)

    num_segs = seg_args['num_segs']
    offset_len = seg_args['offset_len']
    win_len = seg_args['win_len']
    step_len = seg_args['step_len']

    com_len = int(num_segs * step_len + win_len) #combined length of frames

    # convert back to seconds and compute required amount of zero padding
    pad_left = max(0, -offset_len)
    pad_right = max(0, (offset_len + com_len) - file_len)
    load_offset = max(0, offset_len) / rate
    audio_len = int(com_len - pad_left - pad_right)
    duration = audio_len / rate

    # load audio segment
    x, rate = librosa.core.load(path=path, sr=rate, offset=load_offset,\
        duration=duration, mono=False, res_type=resample_method)

    # select channel (for stereo only)
    if np.ndim(x) == 2:
        x = x[channel]

    # check that loaded audio segment has the expected length (give or take 1 sample).
    # if this is not the case, load the entire audio file into memory, then cut out the 
    # relevant section. 
    if abs(len(x) - audio_len) > 1:
        x, rate = librosa.core.load(path=path, sr=rate, mono=False)
        if np.ndim(x) == 2:
            x = x[channel]

        a = max(0, offset_len)
        b = a + audio_len
        x = x[a:b]

    # pad with own reflection
    pad_right += max(0, len(x) - com_len)
    x = aum.pad_reflect(x, pad_left=pad_left, pad_right=pad_right)

    # parse filename
    if id is None: id = os.path.basename(path)

    # create Waveform object
    audio = Waveform(data=x, rate=rate, filename=id, offset=offset)

    return audio, seg_args

class Spectrogram(BaseAudio):
    """ Spectrogram.

        Parent class for MagSpectrogram, PowerSpectrogram, MelSpectrogram, 
        and CQTSpectrogram.

        The Spectrogram class stores the spectrogram pixel values in a 2d 
        numpy array, where the first axis (0) is the time dimension and 
        the second axis (1) is the frequency dimensions.

        The Spectrogram class can also store a stack of multiple, identical-size, 
        spectrograms in a 3d numpy array with the last axis (3) representing the 
        multiple instances.

        Args:
            image: 2d or 3d numpy array
                Spectrogram pixel values. 
            time_res: float
                Time resolution in seconds (corresponds to the bin size used on the time axis)
            spec_type: str
                Spectrogram type. Options include,
                    * 'Mag': Magnitude spectrogram
                    * 'Pow': Power spectrogram
                    * 'Mel': Mel spectrogram
                    * 'CQT': CQT spectrogram
            freq_ax: LinearAxis or Log2Axis
                Axis object for the frequency dimension
            filename: str or list(str)
                Name of the source audio file, if available.   
            offset: float or array-like
                Position in seconds of the left edge of the spectrogram within the source 
                audio file, if available.
            label: int
                Spectrogram label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional
            
        Attributes:
            image: 2d or 3d numpy array
                Spectrogram pixel values. 
            time_ax: LinearAxis
                Axis object for the time dimension
            freq_ax: LinearAxis or Log2Axis
                Axis object for the frequency dimension
            type: str
                Spectrogram type. Options include,
                    * 'Mag': Magnitude spectrogram
                    * 'Pow': Power spectrogram
                    * 'Mel': Mel spectrogram
                    * 'CQT': CQT spectrogram
            filename: str or list(str)
                Name of the source audio file.   
            offset: float or array-like
                Position in seconds of the left edge of the spectrogram within the source 
                audio file.
            label: int
                Spectrogram label.
            annot: AnnotationHandler
                AnnotationHandler object.
"""
    def __init__(self, data, time_res, spec_type, freq_ax, filename=None, offset=0, label=None, annot=None):
        super().__init__(data=data, time_res=time_res, ndim=2, filename=filename, offset=offset, label=label, annot=annot)

        assert freq_ax.bins == data.shape[1], 'data and freq_ax have incompatible shapes'
        self.freq_ax = freq_ax
        self.type = spec_type

    def get_attrs(self):
        return {'time_res':self.time_res(), 'spec_type':self.type, 'freq_ax':self.freq_ax}

    def freq_min(self):
        """ Get spectrogram minimum frequency in Hz.

            Returns:
                : float
                    Frequency in Hz
        """
        return self.freq_ax.min()

    def freq_max(self):
        """ Get spectrogram maximum frequency in Hz.

            Returns:
                : float
                    Frequency in Hz
        """
        return self.freq_ax.max()

    def crop(self, start=None, end=None, length=None,\
        freq_min=None, freq_max=None, height=None, make_copy=False):
        """ Crop spectogram along time axis, frequency axis, or both.
            
            Args:
                start: float
                    Start time in seconds, measured from the left edge of spectrogram.
                end: float
                    End time in seconds, measured from the left edge of spectrogram.
                length: int
                    Horizontal size of the cropped image (number of pixels). If provided, 
                    the `end` argument is ignored. 
                freq_min: float
                    Lower frequency in Hz.
                freq_max: str or float
                    Upper frequency in Hz.
                height: int
                    Vertical size of the cropped image (number of pixels). If provided, 
                    the `freq_max` argument is ignored. 
                make_copy: bool
                    Return a cropped copy of the spectrogra. Leaves the present instance 
                    unaffected. Default is False.

            Returns:
                spec: Spectrogram
                    Cropped spectrogram

            Examples: 
                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from ketos.audio.spectrogram import Spectrogram
                >>> from ketos.audio.utils.axis import LinearAxis
                >>> # Create a spectrogram with shape (20,30), time resolution of 
                >>> # 0.5 s, random pixel values, and a linear frequency axis from 
                >>> # 0 to 300 Hz,
                >>> ax = LinearAxis(bins=30, extent=(0.,300.), label='Frequency (Hz)')
                >>> img = np.random.rand(20,30)
                >>> spec = Spectrogram(data=img, time_res=0.5, spec_type='Mag', freq_ax=ax)
                >>> # Draw the spectrogram
                >>> fig = spec.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/spec_orig.png")
                >>> plt.close(fig)
                
                .. image:: ../../../../ketos/tests/assets/tmp/spec_orig.png

                >>> # Crop the spectrogram along time axis
                >>> spec1 = spec.crop(start=2.0, end=4.2, make_copy=True)
                >>> # Draw the spectrogram
                >>> fig = spec.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/spec_cropped.png")
                >>> plt.close(fig)
                
                .. image:: ../../../../ketos/tests/assets/tmp/spec_cropped.png
        """
        spec = super().crop(start=start, end=end, length=length, make_copy=make_copy) #crop time axis

        # crop frequency axis
        b1, b2 = self.freq_ax.cut(x_min=freq_min, x_max=freq_max, bins=height)

        # crop image
        spec.data = spec.data[:, b1:b2+1]

        # crop annotations, if any
        if self.annot:
            self.annot.crop(freq_min=freq_min, freq_max=freq_max)

        return spec

    def segment(self, window, step=None):
        """ Divide the time axis into segments of uniform length, which may or may 
            not be overlapping.

            Window length and step size are converted to the nearest integer number 
            of time steps.

            If necessary, the spectrogram will be padded with zeros at the end to 
            ensure that all segments have an equal number of samples. 

            Args:
                window: float
                    Length of each segment in seconds.
                step: float
                    Step size in seconds.

            Returns:
                specs: Spectrogram
                    Stacked spectrograms
        """
        segs, filename, offset, label, annot = segment_data(self, window, step)

        ax = copy.deepcopy(self.freq_ax)
        specs = self.__class__(data=segs, time_res=self.time_res(), spec_type=self.type, freq_ax=ax,\
            filename=filename, offset=offset, label=label, annot=annot)
        
        return specs
                
    def add(self, spec, offset=0, scale=1, make_copy=False):
        """ Add another spectrogram on top of this spectrogram.

            The spectrograms must be of the same type, and share the same 
            time resolution. 
            
            The spectrograms must have consistent frequency axes. 
            For linear frequency axes, this implies having the same 
            resolution; for logarithmic axes with base 2, this implies having 
            the same number of bins per octave minimum values that differ by 
            a factor of :math:`2^{n/m}` where :math:`m` is the number of bins 
            per octave and :math:`n` is any integer. No check is made for the 
            consistency of the frequency axes.

            Note that the attributes filename, offset, and label of the spectrogram 
            that is being added are lost.

            The sum spectrogram has the same dimensions (time x frequency) as 
            the original spectrogram.

            Args:
                spec: Spectrogram
                    Spectrogram to be added
                offset: float
                    Shift the spectrograms that is being added by this many seconds 
                    relative to the original spectrogram.
                scale: float
                    Scaling factor applied to spectrogram that is added
                make_copy: bool
                    Make copies of both spectrograms so as to leave the original 
                    instances unchanged.

            Returns:
                : Spectrogram
                    Sum spectrogram
        """
        return add_specs(a=self, b=spec, offset=offset, scale=scale, make_copy=make_copy)

    def blur(self, sigma_time, sigma_freq=0):
        """ Blur the spectrogram using a Gaussian filter.

            Note that the spectrogram frequency axis must be linear if sigma_freq > 0.

            This uses the Gaussian filter method from the scipy.ndimage package:
            
                https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.gaussian_filter.html

            Args:
                sigma_time: float
                    Gaussian kernel standard deviation along time axis in seconds. 
                    Must be strictly positive.
                sigma_freq: float
                    Gaussian kernel standard deviation along frequency axis in Hz.

            Example:        
                >>> from ketos.audio.spectrogram import Spectrogram
                >>> from ketos.audio.waveform import Waveform
                >>> import matplotlib.pyplot as plt
                >>> # create audio signal
                >>> s = Waveform.morlet(rate=1000, frequency=300, width=1)
                >>> # create spectrogram
                >>> spec = MagSpectrogram.from_waveform(s, window=0.2, step=0.05)
                >>> # show image
                >>> fig = spec.plot()
                >>> plt.close(fig)
                >>> # apply very small amount (0.01 sec) of horizontal blur
                >>> # and significant amount of vertical blur (30 Hz)  
                >>> spec.blur(sigma_time=0.01, sigma_freq=30)
                >>> # show blurred image
                >>> fig = spec.plot()
                >>> plt.close(fig)
                
                .. image:: ../../_static/morlet_spectrogram.png

                .. image:: ../../_static/morlet_spectrogram_blurred.png
        """
        assert sigma_time > 0, "sigma_time must be strictly positive"
        sig_t = sigma_time / self.time_res()

        if sigma_freq > 0:
            assert isinstance(self.freq_ax, LinearAxis), "Frequency axis must be linear when sigma_freq > 0"
            sig_f = sigma_freq / self.freq_ax.bin_width()
        else:
            sig_f = 0

        self.data = ndimage.gaussian_filter(input=self.data, sigma=(sig_t, sig_f))

    def enhance_signal(self, enhancement=1.):
        """ Enhance the contrast between regions of high and low intensity.

            See :func:`audio.image.enhance_image` for implementation details.

            Args:
                enhancement: float
                    Parameter determining the amount of enhancement.
        """
        self.data = enhance_signal(self.data, enhancement=enhancement)

    def reduce_tonal_noise(self, method='MEDIAN', **kwargs):
        """ Reduce continuous tonal noise produced by e.g. ships and slowly varying 
            background noise

            See :func:`audio.image.reduce_tonal_noise` for implementation details.

            Currently, offers the following two methods:

                1. MEDIAN: Subtracts from each row the median value of that row.
                
                2. RUNNING_MEAN: Subtracts from each row the running mean of that row.
                
            The running mean is computed according to the formula given in 
            Baumgartner & Mussoline, JASA 129, 2889 (2011); doi: 10.1121/1.3562166

            Args:
                method: str
                    Options are 'MEDIAN' and 'RUNNING_MEAN'
            
            Optional args:
                time_constant: float
                    Time constant in seconds, used for the computation of the running mean.
                    Must be provided if the method 'RUNNING_MEAN' is chosen.

            Example:
                >>> # read audio file
                >>> from ketos.audio.waveform import Waveform
                >>> aud = Waveform.from_wav('ketos/tests/assets/grunt1.wav')
                >>> # compute the spectrogram
                >>> from ketos.audio.spectrogram import MagSpectrogram
                >>> spec = MagSpectrogram.from_waveform(aud, window=0.2, step=0.02)
                >>> # keep only frequencies below 800 Hz
                >>> spec = spec.crop(freq_max=800)
                >>> # show spectrogram as is
                >>> fig = spec.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/spec_before_tonal.png")
                >>> plt.close(fig)
                >>> # tonal noise reduction
                >>> spec.reduce_tonal_noise()
                >>> # show modified spectrogram
                >>> fig = spec.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/spec_after_tonal.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/spec_before_tonal.png

                .. image:: ../../../../ketos/tests/assets/tmp/spec_after_tonal.png

        """
        if 'time_constant' in kwargs.keys():
            time_const_len = kwargs['time_constant'] / self.time_ax.bin_width()
        else:
            time_const_len = None

        self.data = reduce_tonal_noise(self.data, method=method, time_const_len=time_const_len)

    def plot(self, id=0, show_annot=False, figsize=(5,4), cmap='viridis', label_in_title=True):
        """ Plot the spectrogram with proper axes ranges and labels.

            Optionally, also display annotations as boxes superimposed on the spectrogram.

            The colormaps available can be seen here: https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html

            Note: The resulting figure can be shown (fig.show())
            or saved (fig.savefig(file_name))

            Args:
                id: int
                    Spectrogram to be plotted. Only relevant if the spectrogram object 
                    contains multiple, stacked spectrograms.
                show_annot: bool
                    Display annotations
                figsize: tuple
                    Figure size
                cmap: string
                    The colormap to be used
                label_in_title: bool
                    Include label (if available) in figure title
            
            Returns:
                fig: matplotlib.figure.Figure
                A figure object.

            Example:
                >>> from ketos.audio.spectrogram import MagSpectrogram
                >>> # load spectrogram
                >>> spec = MagSpectrogram.from_wav('ketos/tests/assets/grunt1.wav', window=0.2, step=0.02)
                >>> # add an annotation
                >>> spec.annotate(start=1.2, end=1.6, freq_min=70, freq_max=600, label=1)
                >>> # keep only frequencies below 800 Hz
                >>> spec = spec.crop(freq_max=800)
                >>> # show spectrogram with annotation box
                >>> fig = spec.plot(show_annot=True)
                >>> fig.savefig("ketos/tests/assets/tmp/spec_w_annot_box.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/spec_w_annot_box.png
        """
        fig, ax = super().plot(id, figsize, label_in_title)

        x = self.get_data(id) # select image data        
        extent = (0., self.duration(), self.freq_min(), self.freq_max()) # axes ranges        
        img = ax.imshow(x.T, aspect='auto', origin='lower', cmap=cmap, extent=extent)# draw image
        ax.set_ylabel(self.freq_ax.label) # axis label        
        fig.colorbar(img, ax=ax, format='%+2.0f dB')# colobar

        # superimpose annotation boxes
        if show_annot: self._draw_annot_boxes(ax,id)
            
        #fig.tight_layout()
        return fig

    def _draw_annot_boxes(self, ax, id=0):
        """Draws annotations boxes on top of the spectrogram

            Args:
                ax: matplotlib.axes.Axes
                    Axes object
                id: int
                    Object ID
        """
        annots = self.get_annotations(id=id)
        if annots is None: return
        y1 = self.freq_min()
        y2 = self.freq_max()
        for idx,annot in annots.iterrows():
            x1 = annot['start']
            x2 = annot['end']
            if not np.isnan(annot['freq_min']): y1 = annot['freq_min']
            if not np.isnan(annot['freq_max']): y2 = annot['freq_max']
            box = patches.Rectangle((x1,y1),x2-x1,y2-y1,linewidth=1,edgecolor='C1',facecolor='none')
            ax.add_patch(box)
            ax.text(x1, y2, int(annot['label']), ha='left', va='bottom', color='C1')

class MagSpectrogram(Spectrogram):
    """ Magnitude Spectrogram.
    
        Args:
            data: 2d or 3d numpy array
                Spectrogram pixel values. 
            time_res: float
                Time resolution in seconds (corresponds to the bin size used on the time axis)
            freq_min: float
                Lower value of the frequency axis in Hz
            freq_res: float
                Frequency resolution in Hz (corresponds to the bin size used on the frequency axis)
            window_func: str
                Window function used for computing the spectrogram
            filename: str or list(str)
                Name of the source audio file, if available.   
            offset: float or array-like
                Position in seconds of the left edge of the spectrogram within the source 
                audio file, if available.
            label: int
                Spectrogram label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional

        Attrs:
            window_func: str
                Window function.
    """
    def __init__(self, data, time_res, freq_min, freq_res, window_func=None, 
        filename=None, offset=0, label=None, annot=None, **kwargs):

        # create frequency axis
        freq_bins = data.shape[1]
        freq_max  = freq_min + freq_bins * freq_res
        ax = LinearAxis(bins=freq_bins, extent=(freq_min, freq_max), label='Frequency (Hz)')

        # create spectrogram
        super().__init__(data=data, time_res=time_res, spec_type=self.__class__.__name__, freq_ax=ax,
            filename=filename, offset=offset, label=label, annot=annot)

        self.window_func = window_func

    @classmethod
    def from_waveform(cls, audio, window=None, step=None, seg_args=None, window_func='hamming', 
        freq_min=None, freq_max=None):
        """ Create a Magnitude Spectrogram from an :class:`audio_signal.Waveform` by 
            computing the Short Time Fourier Transform (STFT).
        
            Args:
                audio: Waveform
                    Audio signal 
                window: float
                    Window length in seconds
                step: float
                    Step size in seconds
                seg_args: dict
                    Input arguments used for evaluating :func:`audio.audio.segment_args`. 
                    Optional. If specified, the arguments `window` and `step` are ignored.
                window_func: str
                    Window function (optional). Select between
                        * bartlett
                        * blackman
                        * hamming (default)
                        * hanning
                freq_min: float
                    Lower frequency in Hz.
                freq_max: str or float
                    Upper frequency in Hz.

            Returns:
                spec: MagSpectrogram
                    Magnitude spectrogram
        """
        if window_func is not None: window_func = window_func.lower() #make lowercase

        # compute STFT
        img, freq_nyquist, num_fft, seg_args = aum.stft(x=audio.data, rate=audio.rate, window=window,
            step=step, seg_args=seg_args, window_func=window_func)

        time_res = seg_args['step_len'] / audio.rate
        freq_res = freq_nyquist / img.shape[1]

        spec = cls(data=img, time_res=time_res, freq_min=0, freq_res=freq_res, window_func=window_func, 
            filename=audio.filename, offset=audio.offset, label=audio.label, annot=audio.annot)

        spec = spec.crop(freq_min=freq_min, freq_max=freq_max)

        return spec

    @classmethod
    def from_wav(cls, path, window, step, channel=0, rate=None,
            window_func='hamming', offset=0, duration=None,
            resample_method='scipy', freq_min=None, freq_max=None,
            id=None):
        """ Create magnitude spectrogram directly from wav file.

            The arguments offset and duration can be used to select a portion of the wav file.
            
            Note that values specified for the arguments window, step, offset, and duration 
            may all be subject to slight adjustments to ensure that the selected portion 
            corresponds to an integer number of window frames, and that the window and step 
            sizes correspond to an integer number of samples.

            Args:
                path: str
                    Path to wav file
                window: float
                    Window size in seconds
                step: float
                    Step size in seconds 
                channel: int
                    Channel to read from. Only relevant for stereo recordings
                rate: float
                    Desired sampling rate in Hz. If None, the original sampling rate will be used
                window_func: str
                    Window function (optional). Select between
                        * bartlett
                        * blackman
                        * hamming (default)
                        * hanning
                offset: float
                    Start time of spectrogram in seconds, relative the start of the wav file.
                duration: float
                    Length of spectrogram in seconds.
                resample_method: str
                    Resampling method. Only relevant if `rate` is specified. Options are
                        * kaiser_best
                        * kaiser_fast
                        * scipy (default)
                        * polyphase
                    See https://librosa.github.io/librosa/generated/librosa.core.resample.html 
                    for details on the individual methods.
                freq_min: float
                    Lower frequency in Hz.
                freq_max: str or float
                    Upper frequency in Hz.
                id: str
                    Unique identifier (optional). If None, the filename will be used.

            Returns:
                : MagSpectrogram
                    Magnitude spectrogram

            Example:
                >>> # load spectrogram from wav file
                >>> from ketos.audio.spectrogram import MagSpectrogram
                >>> spec = MagSpectrogram.from_wav('ketos/tests/assets/grunt1.wav', window=0.2, step=0.01)
                >>> # crop frequency
                >>> spec = spec.crop(freq_min=50, freq_max=800)
                >>> # show
                >>> fig = spec.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/spec_grunt1.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/spec_grunt1.png
        """
        # load audio
        audio, seg_args = load_audio_for_spec(path=path, channel=channel, rate=rate, window=window, step=step,\
            offset=offset, duration=duration, resample_method=resample_method, id=id)

        # compute spectrogram
        return cls.from_waveform(audio=audio, seg_args=seg_args, window_func=window_func, 
            freq_min=freq_min, freq_max=freq_max)

    def get_attrs(self):
        return {'time_res':self.time_res(), 'freq_min':self.freq_min(), 'freq_res':self.freq_res(), 
            'window_func':self.window_func, 'type':self.__class__.__name__}

    def freq_res(self):
        """ Get frequency resolution in Hz.

            Returns:
                : float
                    Frequency resolution in Hz
        """
        return self.freq_ax.bin_width()

    def recover_waveform(self, num_iters=25, phase_angle=0):
        """ Estimate audio signal from magnitude spectrogram.

            Uses :func:`audio.audio.spec2wave`.

            Args:
                num_iters: 
                    Number of iterations to perform.
                phase_angle: 
                    Initial condition for phase.

            Returns:
                : Waveform
                    Audio signal
        """
        mag = aum.from_decibel(self.data) #use linear scale

        # if the frequency axis has been cropped, pad with zeros to ensure that 
        # the spectrogram has the expected shape
        pad_low  = max(0, int(self.freq_min() / self.freq_res()))
        if pad_low > 0:
            mag = np.pad(mag, pad_width=((0,0),(pad_low,0)), mode='constant')

        target_rate = self.freq_ax.bin_width() * 2 * mag.shape[1]

        # retrieve settings used for computing STFT
        num_fft = 2 * (mag.shape[1] - 1)
        step_len = int(target_rate * self.time_res())  #self.seg_args['step_len']
        if self.window_func:
            window_func = get_window(self.window_func, num_fft)
        else:
            window_func = np.ones(num_fft)

        # iteratively estimate audio signal
        audio = aum.spec2wave(image=mag, phase_angle=phase_angle, num_fft=num_fft,\
            step_len=step_len, num_iters=num_iters, window_func=window_func)

        # sampling rate of recovered audio signal
        rate = len(audio) / (self.duration() + (num_fft - step_len) / target_rate)
        
        return Waveform(rate=rate, data=audio)

class PowerSpectrogram(Spectrogram):
    """ Power Spectrogram.
    
        Args:
            data: 2d or 3d numpy array
                Spectrogram pixel values. 
            time_res: float
                Time resolution in seconds (corresponds to the bin size used on the time axis)
            freq_min: float
                Lower value of the frequency axis in Hz
            freq_res: float
                Frequency resolution in Hz (corresponds to the bin size used on the frequency axis)
            window_func: str
                Window function used for computing the spectrogram
            filename: str or list(str)
                Name of the source audio file, if available.   
            offset: float or array-like
                Position in seconds of the left edge of the spectrogram within the source 
                audio file, if available.
            label: int
                Spectrogram label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional

        Attrs:
            window_func: str
                Window function.
    """
    def __init__(self, data, time_res, freq_min, freq_res, window_func=None, 
        filename=None, offset=0, label=None, annot=None, **kwargs):

        # create frequency axis
        freq_bins = data.shape[1]
        freq_max  = freq_min + freq_bins * freq_res
        ax = LinearAxis(bins=freq_bins, extent=(freq_min, freq_max), label='Frequency (Hz)')

        # create spectrogram
        super().__init__(data=data, time_res=time_res, spec_type=self.__class__.__name__, freq_ax=ax,
            filename=filename, offset=offset, label=label, annot=annot)

        self.window_func = window_func

    @classmethod
    def from_waveform(cls, audio, window=None, step=None, seg_args=None, window_func='hamming', 
        freq_min=None, freq_max=None):
        """ Create a Power Spectrogram from an :class:`audio_signal.Waveform` by 
            computing the Short Time Fourier Transform (STFT).
        
            Args:
                audio: Waveform
                    Audio signal 
                window: float
                    Window length in seconds
                step: float
                    Step size in seconds
                seg_args: dict
                    Input arguments used for evaluating :func:`audio.audio.segment_args`. 
                    Optional. If specified, the arguments `window` and `step` are ignored.
                window_func: str
                    Window function (optional). Select between
                        * bartlett
                        * blackman
                        * hamming (default)
                        * hanning
                freq_min: float
                    Lower frequency in Hz.
                freq_max: str or float
                    Upper frequency in Hz.

            Returns:
                : MagSpectrogram
                    Magnitude spectrogram
        """
        if window_func is not None: window_func = window_func.lower() #make lowercase

        # compute STFT
        img, freq_nyquist, num_fft, seg_args = aum.stft(x=audio.data, rate=audio.rate, window=window,\
            step=step, seg_args=seg_args, window_func=window_func, decibel=False)
        img = mag2pow(img, num_fft) # Magnitude->Power conversion
        img = aum.to_decibel(img) # convert to dB

        time_res = seg_args['step_len'] / audio.rate
        freq_res = freq_nyquist / img.shape[1]

        spec = cls(data=img, time_res=time_res, freq_min=0, freq_res=freq_res, window_func=window_func, 
            filename=audio.filename, offset=audio.offset, label=audio.label, annot=audio.annot)

        spec = spec.crop(freq_min=freq_min, freq_max=freq_max)

        return spec

    @classmethod
    def from_wav(cls, path, window, step, channel=0, rate=None,
            window_func='hamming', offset=0, duration=None,
            resample_method='scipy', freq_min=None, freq_max=None,
            id=None):            
        """ Create power spectrogram directly from wav file.

            The arguments offset and duration can be used to select a portion of the wav file.
            
            Note that values specified for the arguments window, step, offset, and duration 
            may all be subject to slight adjustments to ensure that the selected portion 
            corresponds to an integer number of window frames, and that the window and step 
            sizes correspond to an integer number of samples.

            Args:
                path: str
                    Path to wav file
                window: float
                    Window size in seconds
                step: float
                    Step size in seconds 
                channel: int
                    Channel to read from. Only relevant for stereo recordings
                rate: float
                    Desired sampling rate in Hz. If None, the original sampling rate will be used
                window_func: str
                    Window function (optional). Select between
                        * bartlett
                        * blackman
                        * hamming (default)
                        * hanning
                offset: float
                    Start time of spectrogram in seconds, relative the start of the wav file.
                duration: float
                    Length of spectrogrma in seconds.
                resample_method: str
                    Resampling method. Only relevant if `rate` is specified. Options are
                        * kaiser_best
                        * kaiser_fast
                        * scipy (default)
                        * polyphase
                    See https://librosa.github.io/librosa/generated/librosa.core.resample.html 
                    for details on the individual methods.
                freq_min: float
                    Lower frequency in Hz.
                freq_max: str or float
                    Upper frequency in Hz.
                id: str
                    Unique identifier (optional). If None, the filename will be used.

            Returns:
                spec: MagSpectrogram
                    Magnitude spectrogram

            Example:
                >>> # load spectrogram from wav file
                >>> from ketos.audio.spectrogram import MagSpectrogram
                >>> spec = MagSpectrogram.from_wav('ketos/tests/assets/grunt1.wav', window=0.2, step=0.01)
                >>> # crop frequency
                >>> spec = spec.crop(freq_min=50, freq_max=800)
                >>> # show
                >>> fig = spec.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/spec_grunt1.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/spec_grunt1.png
        """
        # load audio
        audio, seg_args = load_audio_for_spec(path=path, channel=channel, rate=rate, window=window, step=step,\
            offset=offset, duration=duration, resample_method=resample_method, id=id)

        # compute spectrogram
        return cls.from_waveform(audio=audio, seg_args=seg_args, window_func=window_func, 
            freq_min=freq_min, freq_max=freq_max)

    def get_attrs(self):
        return {'time_res':self.time_res(), 'freq_min':self.freq_min(), 'freq_res':self.freq_res(), 
            'window_func':self.window_func, 'type':self.__class__.__name__}

    def freq_res(self):
        """ Get frequency resolution in Hz.

            Returns:
                : float
                    Frequency resolution in Hz
        """
        return self.freq_ax.bin_width()

class MelSpectrogram(Spectrogram):
    """ Mel Spectrogram.
    
        Args:
            data: 2d or 3d numpy array
                Mel spectrogram pixel values. 
            filter_banks: numpy.array
                Filter banks
            time_res: float
                Time resolution in seconds (corresponds to the bin size used on the time axis)
            freq_min: float
                Lower value of the frequency axis in Hz
            freq_max: float
                Upper value of the frequency axis in Hz
            window_func: str
                Window function used for computing the spectrogram
            filename: str or list(str)
                Name of the source audio file, if available.   
            offset: float or array-like
                Position in seconds of the left edge of the spectrogram within the source 
                audio file, if available.
            label: int
                Spectrogram label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional

        Attrs:
            window_func: str
                Window function.
            filter_banks: numpy.array
                Filter banks
    """
    def __init__(self, data, filter_banks, time_res, freq_min, freq_max, 
        window_func=None, filename=None, offset=0, label=None, annot=None, **kwargs):

        # create frequency axis
        # TODO: this needs to be modified as the Mel frequency axis is not linear
        ax = LinearAxis(bins=data.shape[1], extent=(freq_min, freq_max), label='Frequency (Hz)')

        # create spectrogram
        super().__init__(data=data, time_res=time_res, spec_type=self.__class__.__name__, freq_ax=ax,
            filename=filename, offset=offset, label=label, annot=annot)

        self.window_func = window_func
        self.filter_banks = filter_banks

    @classmethod
    def from_waveform(cls, audio, window=None, step=None, seg_args=None, window_func='hamming',
        num_filters=40, num_ceps=20, cep_lifter=20):
        """ Creates a Mel Spectrogram from an :class:`audio_signal.Waveform`.
        
            Args:
                audio: Waveform
                    Audio signal 
                window: float
                    Window length in seconds
                step: float
                    Step size in seconds
                seg_args: dict
                    Input arguments used for evaluating :func:`audio.audio.segment_args`. 
                    Optional. If specified, the arguments `window` and `step` are ignored.
                window_func: str
                    Window function (optional). Select between
                        * bartlett
                        * blackman
                        * hamming (default)
                        * hanning
                num_filters: int
                    The number of filters in the filter bank.
                num_ceps: int
                    The number of Mel-frequency cepstrums.
                cep_lifters: int
                    The number of cepstum filters.

            Returns:
                : MelSpectrogram
                    Mel spectrogram
        """
        if window_func is not None: window_func = window_func.lower() #make lowercase

        # compute STFT
        img, freq_nyquist, num_fft, seg_args = aum.stft(x=audio.data, rate=audio.rate, window=window,\
            step=step, seg_args=seg_args, window_func=window_func, decibel=False)

        # Magnitude->Mel conversion
        img, filter_banks = mag2mel(img=img, num_fft=num_fft, rate=audio.rate,\
            num_filters=num_filters, num_ceps=num_ceps, cep_lifter=cep_lifter) 

        time_res = seg_args['step_len'] / audio.rate   

        return cls(data=img, filter_banks=filter_banks, time_res=time_res, 
            freq_min=0, freq_max=freq_nyquist, window_func=window_func, 
            filename=audio.filename, offset=audio.offset, label=audio.label, annot=audio.annot)

    @classmethod
    def from_wav(cls, path, window, step, channel=0, rate=None,\
            window_func='hamming', num_filters=40, num_ceps=20, cep_lifter=20,\
            offset=0, duration=None, resample_method='scipy',
            id=None):            
        """ Create Mel spectrogram directly from wav file.

            The arguments offset and duration can be used to select a portion of the wav file.
            
            Note that values specified for the arguments window, step, offset, and duration 
            may all be subject to slight adjustments to ensure that the selected portion 
            corresponds to an integer number of window frames, and that the window and step 
            sizes correspond to an integer number of samples.

            Args:
                path: str
                    Path to wav file
                window: float
                    Window size in seconds
                step: float
                    Step size in seconds 
                channel: int
                    Channel to read from. Only relevant for stereo recordings
                rate: float
                    Desired sampling rate in Hz. If None, the original sampling rate will be used
                window_func: str
                    Window function (optional). Select between
                        * bartlett
                        * blackman
                        * hamming (default)
                        * hanning
                num_filters: int
                    The number of filters in the filter bank.
                num_ceps: int
                    The number of Mel-frequency cepstrums.
                cep_lifters: int
                    The number of cepstum filters.
                offset: float
                    Start time of spectrogram in seconds, relative the start of the wav file.
                duration: float
                    Length of spectrogrma in seconds.
                resample_method: str
                    Resampling method. Only relevant if `rate` is specified. Options are
                        * kaiser_best
                        * kaiser_fast
                        * scipy (default)
                        * polyphase
                    See https://librosa.github.io/librosa/generated/librosa.core.resample.html 
                    for details on the individual methods.
                id: str
                    Unique identifier (optional). If None, the filename will be used.

            Returns:
                spec: MelSpectrogram
                    Mel spectrogram

            Example:
                >>> # load spectrogram from wav file
                >>> from ketos.audio.spectrogram import MelSpectrogram
                >>> spec = MelSpectrogram.from_wav('ketos/tests/assets/grunt1.wav', window=0.2, step=0.01)
                >>> # crop frequency
                >>> spec = spec.crop(freq_min=50, freq_max=800)
                >>> # show
                >>> fig = spec.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/mel_grunt1.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/mel_grunt1.png
        """
        # load audio
        audio, seg_args = load_audio_for_spec(path=path, channel=channel, rate=rate, window=window, step=step,\
            offset=offset, duration=duration, resample_method=resample_method, id=id)

        # compute spectrogram
        spec = cls.from_waveform(audio=audio, seg_args=seg_args, window_func=window_func, 
            num_filters=num_filters, num_ceps=num_ceps, cep_lifter=cep_lifter)

        return spec

    def get_attrs(self):
        return {'time_res':self.time_res(), 'freq_min':self.freq_min(), 'freq_max':self.freq_max(), 
            'filter_banks':self.filter_banks, 'type':self.__class__.__name__}

    def plot(self, filter_bank=False, figsize=(5,4), cmap='viridis'):
        """ Plot the spectrogram with proper axes ranges and labels.

            The colormaps available can be seen here: https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html

            Note: The resulting figure can be shown (fig.show())
            or saved (fig.savefig(file_name))

            TODO: Check implementation for filter_bank=True

            Args:
                filter_bank: bool
                    If True, plot the filter banks if True. If False (default), 
                    print the mel spectrogram.
                figsize: tuple
                    Figure size
                cmap: string
                    The colormap to be used
            
            Returns:
                fig: matplotlib.figure.Figure
                    A figure object.
        """
        if filter_bank:
            img = self.filter_banks
            fig, ax = plt.subplots(figsize=figsize)
            extent = (0,self.duration(),self.freq_min(),self.freq_max())
            img_plot = ax.imshow(img.T,aspect='auto',origin='lower', cmap=cmap, extent=extent)
            ax.set_xlabel(self.time_ax.label)
            fig.colorbar(img_plot,format='%+2.0f dB')

        else:
            fig = super().plot(cmap=cmap, figsize=figsize)

        return fig

class CQTSpectrogram(Spectrogram):
    """ Magnitude Spectrogram computed from Constant Q Transform (CQT).
    
        Args:
            image: 2d or 3d numpy array
                Spectrogram pixel values. 
            time_res: float
                Time resolution in seconds (corresponds to the bin size used on the time axis)
            freq_min: float
                Lower value of the frequency axis in Hz
            bins_per_oct: int
                Number of bins per octave
            window_func: str
                Window function used for computing the spectrogram
            filename: str or list(str)
                Name of the source audio file, if available.   
            offset: float or array-like
                Position in seconds of the left edge of the spectrogram within the source 
                audio file, if available.
            label: int
                Spectrogram label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional

        Attrs:
            window_func: str
                Window function.
    """
    def __init__(self, data, time_res, bins_per_oct, freq_min, 
        window_func=None, filename=None, offset=0, label=None, annot=None, **kwargs):

        # create logarithmic frequency axis
        ax = Log2Axis(bins=data.shape[1], bins_per_oct=bins_per_oct,\
            min_value=freq_min, label='Frequency (Hz)')

        # create spectrogram
        super().__init__(data=data, time_res=time_res, spec_type=self.__class__.__name__, freq_ax=ax,
            filename=filename, offset=offset, label=label, annot=annot)

        self.window_func = window_func

    @classmethod
    def from_waveform(cls, audio, step, bins_per_oct, freq_min=1, freq_max=None, window_func='hann'):
        """ Magnitude Spectrogram computed from Constant Q Transform (CQT) using the librosa implementation:

            https://librosa.github.io/librosa/generated/librosa.core.cqt.html

            The frequency axis of a CQT spectrogram is essentially a logarithmic axis with base 2. It is 
            characterized by an integer number of bins per octave (an octave being a doubling of the frequency.) 

            For further details, see :func:`audio.audio.cqt`.
        
            Args:
                audio: Waveform
                    Audio signal 
                step: float
                    Step size in seconds 
                bins_per_oct: int
                    Number of bins per octave
                freq_min: float
                    Minimum frequency in Hz. Default is 1 Hz.
                freq_max: float
                    Maximum frequency in Hz
                    If None, it is set half the sampling rate.
                window_func: str
                    Window function (optional). Select between
                        * bartlett
                        * blackman
                        * hamming
                        * hanning (default)

            Returns:
                spec: CQTSpectrogram
                    CQT spectrogram
        """
        if window_func is not None: window_func = window_func.lower() #make lowercase

        # compute CQT
        img, step = aum.cqt(x=audio.data, rate=audio.rate, step=step,
            bins_per_oct=bins_per_oct, freq_min=freq_min, freq_max=freq_max,
            window_func=window_func)

        spec = cls(data=img, time_res=step, freq_min=freq_min, bins_per_oct=bins_per_oct, 
            window_func=window_func, filename=audio.filename, 
            offset=audio.offset, label=audio.label, annot=audio.annot)

        spec = spec.crop(freq_min=freq_min, freq_max=freq_max)

        return spec

    @classmethod
    def from_wav(cls, path, step, bins_per_oct, freq_min=1, freq_max=None,
        channel=0, rate=None, window_func='hann', offset=0, duration=None,
        resample_method='scipy', id=None):
        """ Create CQT spectrogram directly from wav file.

            The arguments offset and duration can be used to select a segment of the audio file.

            Note that values specified for the arguments window, step, offset, and duration 
            may all be subject to slight adjustments to ensure that the selected portion 
            corresponds to an integer number of window frames, and that the window and step 
            sizes correspond to an integer number of samples.
        
            Args:
                path: str
                    Complete path to wav file 
                step: float
                    Step size in seconds 
                bins_per_oct: int
                    Number of bins per octave
                freq_min: float
                    Minimum frequency in Hz. Default is 1 Hz.
                freq_max: float
                    Maximum frequency in Hz
                    If None, it is set half the sampling rate.
                channel: int
                    Channel to read from. Only relevant for stereo recordings
                rate: float
                    Desired sampling rate in Hz. If None, the original sampling rate will be used
                window_func: str
                    Window function (optional). Select between
                        * bartlett
                        * blackman
                        * hamming (default)
                        * hanning
                offset: float
                    Start time of spectrogram in seconds, relative the start of the wav file.
                duration: float
                    Length of spectrogrma in seconds.
                resample_method: str
                    Resampling method. Only relevant if `rate` is specified. Options are
                        * kaiser_best
                        * kaiser_fast
                        * scipy (default)
                        * polyphase
                    See https://librosa.github.io/librosa/generated/librosa.core.resample.html 
                    for details on the individual methods.
                id: str
                    Unique identifier (optional). If None, the filename will be used.

            Returns:
                : CQTSpectrogram
                    CQT spectrogram

            Example:
                >>> # load spectrogram from wav file
                >>> from ketos.audio.spectrogram import CQTSpectrogram
                >>> spec = CQTSpectrogram.from_wav('ketos/tests/assets/grunt1.wav', step=0.01, freq_min=10, freq_max=800, bins_per_oct=16)
                >>> # show
                >>> fig = spec.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/cqt_grunt1.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/cqt_grunt1.png
        """
        if rate is None:
            rate = librosa.get_samplerate(path) #if not specified, use original sampling rate

        file_duration = librosa.get_duration(filename=path) #get file duration
        file_len = int(file_duration * rate) #file length (number of samples)
        
        assert offset < file_duration, 'Offset exceeds file duration'
        
        if duration is None:
            duration = file_duration - offset #if not specified, use file duration minus offset

        # load audio
        audio = Waveform.from_wav(path=path, rate=rate, channel=channel,
            offset=offset, duration=duration, resample_method=resample_method, id=id)

        # create CQT spectrogram
        return cls.from_waveform(audio=audio, step=step, bins_per_oct=bins_per_oct, 
            freq_min=freq_min, freq_max=freq_max, window_func=window_func)

    def get_attrs(self):
        return {'time_res':self.time_res(), 'freq_min':self.freq_min(), 'bins_per_oct':self.bins_per_octave(), 
            'window_func':self.window_func, 'type':self.__class__.__name__}

    def bins_per_octave(self):
        """ Get no. bins per octave.

            Returns:
                : int
                    No. bins per octave.
        """
        return self.freq_ax.bins_per_oct

    def plot(self, id=0, show_annot=False, figsize=(5,4), cmap='viridis', label_in_title=True):
        """ Plot the spectrogram with proper axes ranges and labels.

            Optionally, also display annotations as boxes superimposed on the spectrogram.

            The colormaps available can be seen here: https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html

            Note: The resulting figure can be shown (fig.show())
            or saved (fig.savefig(file_name))

            Args:
                id: int
                    Spectrogram to be plotted. Only relevant if the spectrogram object 
                    contains multiple, stacked spectrograms.
                show_annot: bool
                    Display annotations
                figsize: tuple
                    Figure size
                cmap: string
                    The colormap to be used
                label_in_title: bool
                    Include label (if available) in figure title
            
            Returns:
                fig: matplotlib.figure.Figure
                    A figure object.
        """
        fig = super().plot(id, show_annot, figsize, cmap, label_in_title)
        ticks, labels = self.freq_ax.ticks_and_labels()
        plt.yticks(ticks, labels)
        return fig



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'utils' module within the ketos library
"""

import pytest
import numpy as np
import pandas as pd
from ketos.utils import tostring, morlet_func, octave_bands, random_floats, nearest_values,\
    detect_peaks, str_is_int


def test_tostring():
    box = (1,2,3)
    s = tostring(box)
    assert s == '[1,2,3]'
    box = [1,2,3]
    s = tostring(box)
    assert s == '[1,2,3]'
    box = np.array([1,2,3])
    s = tostring(box)
    assert s == '[1,2,3]'
    box = [[1,2,3],[1,2]]
    s = tostring(box)
    assert s == '[[1,2,3],[1,2]]'

def test_octave_bands():
    fc, fmin, fmax = octave_bands(1, 3)
    assert fc[0] == 62.5
    assert fc[1] == 125.
    assert fc[2] == 250.

def test_morlet_func_single_time():
    time = 0.5
    f = morlet_func(time=time, frequency=10, width=3, displacement=0)
    assert f == pytest.approx(0.42768108, abs=1E-5) 

def test_morlet_func_multiple_times():
    time = np.array([-1., 0., 0.5])
    f = morlet_func(time=time, frequency=10, width=3, displacement=0)
    assert f[0] == pytest.approx(0.41022718, abs=1E-5) 
    assert f[1] == pytest.approx(0.43366254, abs=1E-5) 
    assert f[2] == pytest.approx(0.42768108, abs=1E-5) 

def test_morlet_func_with_dfdt_nonzero():
    time = 0.5
    f = morlet_func(time=time, frequency=10, width=3, displacement=0, dfdt=0.5)
    assert f == pytest.approx(0.302416, abs=1E-5) 

def test_random_floats():
    x = random_floats(3, 0.4, 7.2)
    assert x[0] == pytest.approx(3.23574963, abs=1e-5)
    assert x[1] == pytest.approx(5.29820656, abs=1e-5)
    assert x[2] == pytest.approx(0.40077775, abs=1e-5)

def test_nearest_values():
    x = np.array([1.0, 4.0, 5.1, 6.0, 0.2, 0.3, 10.0])
    y = nearest_values(x=x, i=3, n=3)
    assert np.all(y == [5.1, 6.0, 0.2])
    y = nearest_values(x=x, i=0, n=3)
    assert np.all(y == [1.0, 4.0, 5.1])
    y = nearest_values(x=x, i=0, n=4)
    assert np.all(y == [1.0, 4.0, 5.1, 6.0])
    y = nearest_values(x=x, i=6, n=4)
    assert np.all(y == [10.0, 0.3, 0.2, 6.0])

def test_detect_peaks():
    # create a two time series, where only the first contains a peak
    d = {'series1' : pd.Series([1.0, 2.3, 22.0, 2.2, 1.5]), 'series2': pd.Series([1.0, 2.3, 1.8, 2.2, 1.5])}
    df = pd.DataFrame(d)
    # detect peaks with multiplicity 1 and prominence of at least 2.0
    peaks = detect_peaks(df=df, multiplicity=1, prominence=2.0)
    assert np.all(peaks == [0, 0, 1, 0, 0])
    # try again, but this time require multiplicity 2
    peaks = detect_peaks(df=df, multiplicity=2, prominence=2.0)
    assert np.all(peaks == [0, 0, 0, 0, 0])

def test_str_is_int():
    assert str_is_int('5')
    assert str_is_int('-5')
    assert str_is_int('+5')
    assert not str_is_int('-5', signed=False)
    assert not str_is_int('5.')



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the selection_table module within the ketos library
"""
import pytest
import os
import ketos.data_handling.selection_table as st
import pandas as pd
import numpy as np
from io import StringIO


current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


def test_trim():
    standard = ['filename','label','start','end','freq_min','freq_max']
    extra = ['A','B','C']
    df = pd.DataFrame(columns=extra)
    df = st.trim(df)
    assert len(df.columns) == 0
    df = pd.DataFrame(columns=standard+extra)
    df = st.trim(df)
    assert sorted(df.columns.values) == sorted(standard)

def test_missing_columns():
    standard = ['filename','label','start','end','freq_min','freq_max']
    df = pd.DataFrame(columns=standard)
    assert len(st.missing_columns(df)) == 0
    df = pd.DataFrame(columns=standard[:-1])
    assert len(st.missing_columns(df)) == 0
    df = pd.DataFrame(columns=standard[1:])
    assert sorted(st.missing_columns(df)) == ['filename']

def test_is_standardized():
    df = pd.DataFrame({'filename':'test.wav','label':[1],'start':[0],'end':[2],'freq_min':[None],'freq_max':[None]})
    df = st.standardize(df)
    assert st.is_standardized(df) == True
    df = pd.DataFrame({'filename':'test.wav','label':[1]})
    df = st.standardize(df)
    assert st.is_standardized(df) == True
    df = pd.DataFrame({'filename':'test.wav','label':[1]})
    assert st.is_standardized(df) == False

def test_create_label_dict():
    l1 = [0, 'gg', -17, 'whale']
    l2 = [-33, 1, 'boat']
    l3 = [999]
    d = st.create_label_dict(l1, l2, l3)
    ans = {-33: 0, 1:0, 'boat': 0, 999: -1, 0: 1, 'gg':2, -17: 3, 'whale': 4}
    assert d == ans

def test_create_label_dict_can_handle_nested_list():
    l1 = [0, 'gg', [-17, 'whale']]
    l2 = [-33, 1, 'boat']
    l3 = [999]
    d = st.create_label_dict(l1, l2, l3)
    ans = {-33: 0, 1:0, 'boat': 0, 999: -1, 0: 1, 'gg':2, -17: 3, 'whale': 3}
    assert d == ans

def test_unfold(annot_table_mult_labels):
    res = st.unfold(annot_table_mult_labels)
    ans = pd.DataFrame({'filename':['f0.wav','f0.wav','f1.wav'], 'label':['1','2','3'], 'start':[0,0,1], 'end':[1,1,2]})
    res = res.reset_index(drop=True)[ans.columns]
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])

def test_standardize(annot_table_std):
    res = st.standardize(annot_table_std)
    d = '''filename annot_id label  start  end                   
f0.wav   0             3    0.0  3.3
f0.wav   1             2    3.0  6.3
f1.wav   0             4    1.0  4.3
f1.wav   1             2    4.0  7.3
f2.wav   0             5    2.0  5.3
f2.wav   1             1    5.0  8.3'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])

def test_standardize_from_file(annot_table_file):
    res, d = st.standardize(filename=annot_table_file, mapper={'fname': 'filename', 'STOP': 'end'}, 
        signal_labels=[1,'k'], backgr_labels=[-99, 'whale'], return_label_dict=True)
    ans = {-99: 0, 'whale':0, 2: -1, 'zebra': -1, 1: 1, 'k':2}
    assert d == ans
    d = '''filename annot_id label  start  end                   
f0.wav   0             1      0    1
f1.wav   0            -1      1    2
f2.wav   0             2      2    3
f3.wav   0             0      3    4
f4.wav   0             0      4    5
f5.wav   0            -1      5    6'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])

def test_standardize_with_nested_list(annot_table_file):
    res, d = st.standardize(filename=annot_table_file, mapper={'fname': 'filename', 'STOP': 'end'}, 
        signal_labels=[[1,'whale'],'k'], backgr_labels=[-99], return_label_dict=True)
    ans = {-99: 0, 2: -1, 'zebra': -1, 1: 1, 'whale':1, 'k':2}
    assert d == ans
    d = '''filename annot_id label  start  end                   
f0.wav   0             1      0    1
f1.wav   0            -1      1    2
f2.wav   0             2      2    3
f3.wav   0             0      3    4
f4.wav   0             1      4    5
f5.wav   0            -1      5    6'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])
    
def test_label_occurrence(annot_table_std):
    df = annot_table_std
    oc = st.label_occurrence(df)
    ans = {-1: 1, 0: 2, 1: 1, 2: 1, 3: 1}
    assert oc == ans

def test_select_center(annot_table_std):
    df = st.standardize(annot_table_std)
    # request length shorter than annotations
    res = st.select(df, length=1, center=True)
    d = '''filename sel_id label  start   end
f0.wav   0           3   1.15  2.15
f0.wav   1           2   4.15  5.15
f1.wav   0           4   2.15  3.15
f1.wav   1           2   5.15  6.15
f2.wav   0           5   3.15  4.15
f2.wav   1           1   6.15  7.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])
    # request length longer than annotations
    res = st.select(df, length=5, center=True)
    d = '''filename sel_id  label  start   end
f0.wav   0           3  -0.85  4.15
f0.wav   1           2   2.15  7.15
f1.wav   0           4   0.15  5.15
f1.wav   1           2   3.15  8.15
f2.wav   0           5   1.15  6.15
f2.wav   1           1   4.15  9.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])

def test_select_removes_discarded_annotations(annot_table_std):
    df = annot_table_std
    df = st.standardize(df)
    res = st.select(df, length=1, center=True)
    assert len(res[res.label==-1]) == 0

def test_select_enforces_overlap(annot_table_std):
    np.random.seed(3)
    df = annot_table_std
    df = st.standardize(df)
    # requested length: 5.0 sec
    # all annotations have length: 3.3 sec  (3.3/5.0=0.66)
    length = 5.0
    overlap = 0.5
    df_new = st.select(df, length=length, min_overlap=overlap, keep_id=True)
    t1 = df_new.start.values
    t2 = df_new.end.values
    idx = zip(df_new.index.get_level_values(0), df_new.annot_id)
    df = df.loc[idx]
    t2_orig = df.end.values
    t1_orig = df.start.values
    assert np.all(t2 >= t1_orig + overlap * length)
    assert np.all(t1 <= t2_orig - overlap * length)

def test_select_step(annot_table_std):
    df = annot_table_std
    df = st.standardize(df)
    N = len(df[df['label']!=-1])
    K = len(df[df['label']==0])
    df_new = st.select(df, length=1, center=True, min_overlap=0, step=0.5, keep_id=True)
    M = len(df_new)
    assert M == (N - K) * (2 * int((3.3/2+0.5)/0.5) + 1) + K * (2 * int((3.3/2-0.5)/0.5) + 1)
    df_new = st.select(df, length=1, center=True, min_overlap=0.4, step=0.5)
    M = len(df_new)
    assert M == (N - K) * (2 * int((3.3/2+0.5-0.4)/0.5) + 1) + K * (2 * int((3.3/2-0.5)/0.5) + 1)

def test_time_shift(annot_table_std):
    row = pd.Series({'label':3.00,'start':0.00,'end':3.30,'annot_id':0.00,'length':3.30,'start_new':-0.35})
    res = st.time_shift(annot=row, time_ref=row['start_new'], length=4.0, min_overlap=0.8, step=0.5)
    d = '''label  start  end  annot_id  length  start_new
0    3.0    0.0  3.3       0.0     3.3      -0.35'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0])
    pd.testing.assert_frame_equal(ans, res[ans.columns.values])

def test_select_with_varying_overlap(annot_table_std):
    """ Test that the number of selections increases as the 
        minimum required overlap is reduced"""
    df = st.standardize(annot_table_std)
    # request length shorter than annotations
    num_sel = []
    for min_overlap in np.linspace(1.0, 0.0, 11):
        res = st.select(df, length=1.0, step=0.5, center=True, min_overlap=min_overlap)
        num_sel.append(len(res))
    
    assert np.all(np.diff(num_sel) >= 0)
    # request length longer than annotations
    num_sel = []
    for min_overlap in np.linspace(1.0, 0.0, 11):
        res = st.select(df, length=4.0, step=0.5, center=True, min_overlap=min_overlap)
        num_sel.append(len(res))

    assert np.all(np.diff(num_sel) >= 0)

def test_create_rndm_backgr_selections(annot_table_std, file_duration_table):
    np.random.seed(1)
    df = st.standardize(annot_table_std)
    dur = file_duration_table 
    num = 5
    df_bgr = st.create_rndm_backgr_selections(annotations=df, files=dur, length=2.0, num=num)
    assert len(df_bgr) == num
    # assert selections have uniform length
    assert np.all(df_bgr.end.values - df_bgr.start.values == 2.0)
    # assert all selection have label = 0
    assert np.all(df_bgr.label.values == 0)
    # assert selections do not overlap with any annotations
    for bgr_idx, bgr_sel in df_bgr.iterrows():
        start_bgr = bgr_sel.start
        end_bgr = bgr_sel.end
        fname = bgr_idx[0]
        q = st.query(df, start=start_bgr, end=end_bgr, filename=fname)
        assert len(q) == 0

def test_create_rndm_backgr_keeps_misc_cols(annot_table_std, file_duration_table):
    """ Check that the random background selection creation method keeps 
        any miscellaneous columns"""
    np.random.seed(1)
    df = st.standardize(annot_table_std)
    dur = file_duration_table 
    dur['extra'] = 'testing'
    df_bgr = st.create_rndm_backgr_selections(annotations=df, files=dur, length=2.0, num=5)
    assert np.all(df_bgr['extra'].values == 'testing')
    df_bgr = st.create_rndm_backgr_selections(annotations=df, files=dur, length=2.0, num=5, trim_table=True)
    assert 'extra' not in df_bgr.columns.values.tolist()

def test_create_rndm_backgr_files_missing_duration(annot_table_std, file_duration_table):
    """ Check that the random background selection creation method works even when 
        some of the files are missing from the file duration list"""
    np.random.seed(1)
    df = st.standardize(annot_table_std)
    dur = file_duration_table.drop(0) 
    df_bgr = st.create_rndm_backgr_selections(annotations=df, files=dur, length=2.0, num=11)

def test_create_rndm_backgr_nonzero_offset(annot_table_std, file_duration_table):
    """ Check that the random background selection creation method works when file duration 
        table includes offsets"""
    np.random.seed(1)
    df = st.standardize(annot_table_std)
    file_duration_table['offset'] = [1, 2, 3, 4, 5, 6]
    df_bgr = st.create_rndm_backgr_selections(annotations=df, files=file_duration_table, length=2.0, num=11)

def test_create_rndm_backgr_selections_no_overlap(annot_table_std, file_duration_table):
    """ Check that random selections have no overlap"""
    np.random.seed(1)
    df = st.standardize(annot_table_std)
    dur = file_duration_table 
    num = 30
    df_bgr = st.create_rndm_backgr_selections(annotations=df, files=dur, length=2.0, num=num)
    num_overlap = 0
    for idx,row in df_bgr.iterrows():
        q = st.query(df_bgr, filename=idx[0], start=row['start'], end=row['end'])
        num_overlap += len(q) - 1
    
    assert num_overlap > 0

    df_bgr = st.create_rndm_backgr_selections(annotations=df, files=dur, length=2.0, num=num, no_overlap=True)
    num_overlap = 0
    for idx,row in df_bgr.iterrows():
        q = st.query(df_bgr, filename=idx[0], start=row['start'], end=row['end'])
        num_overlap += len(q) - 1
    
    assert num_overlap == 0

def test_select_by_segmenting(annot_table_std, file_duration_table):
    a = st.standardize(annot_table_std)
    f = file_duration_table
    sel = st.select_by_segmenting(f, length=5.1, annotations=a, step=4.0, discard_empty=True, pad=True)
    # check selection table
    d = '''filename sel_id start  end
f0.wav   0         0.0  5.1
f0.wav   1         4.0  9.1
f1.wav   0         0.0  5.1
f1.wav   1         4.0  9.1
f2.wav   0         0.0  5.1
f2.wav   1         4.0  9.1
f2.wav   2         8.0 13.1'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(ans, sel[0][ans.columns.values])
    # check annotation table
    d = '''filename sel_id annot_id label  start  end
f0.wav   0      0             3         0.0        3.3
f0.wav   0      1             2         3.0        5.1
f0.wav   1      1             2         0.0        2.3
f1.wav   0      0             4         1.0        4.3
f1.wav   0      1             2         4.0        5.1
f1.wav   1      0             4         0.0        0.3
f1.wav   1      1             2         0.0        3.3
f2.wav   0      0             5         2.0        5.1
f2.wav   0      1             1         5.0        5.1
f2.wav   1      0             5         0.0        1.3
f2.wav   1      1             1         1.0        4.3
f2.wav   2      1             1         0.0        0.3'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1,2])
    pd.testing.assert_frame_equal(ans, sel[1][ans.columns.values])

def test_query_labeled(annot_table_std):
    df = st.standardize(annot_table_std)
    df = st.select(df, length=1, center=True)
    # query for file that does not exist
    q = st.query_labeled(df, filename='fff.wav')
    assert len(q) == 0
    # query for 1 file
    q = st.query_labeled(df, filename='f1.wav')
    d = '''filename sel_id label  start   end                   
f1.wav  0           4   2.15  3.15
f1.wav  1           2   5.15  6.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(q, ans[q.columns.values])
    # query for 1 file, and 1 that does not exist
    q = st.query_labeled(df, filename=['f1.wav','fff.wav'])
    d = '''filename sel_id label  start   end                   
f1.wav   0           4   2.15  3.15
f1.wav   1           2   5.15  6.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(q, ans[q.columns.values])
    # query for 2 files
    q = st.query_labeled(df, filename=['f1.wav','f2.wav'])
    d = '''filename sel_id label  start   end                            
f1.wav   0           4   2.15  3.15
f1.wav   1           2   5.15  6.15
f2.wav   0           5   3.15  4.15
f2.wav   1           1   6.15  7.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(q, ans[q.columns.values])
    # query for labels
    q = st.query_labeled(df, label=[2,5])
    d = '''filename sel_id label  start   end                   
f0.wav   1           2   4.15  5.15
f1.wav   1           2   5.15  6.15
f2.wav   0           5   3.15  4.15'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(q, ans[q.columns.values])
    # query for label that does not exist
    q = st.query_labeled(df, label=99)
    assert len(q) == 0

def test_query_annotated(annot_table_std, file_duration_table):
    a = st.standardize(annot_table_std)
    f = file_duration_table
    sel = st.select_by_segmenting(f, length=5.1, annotations=a, step=4.0, discard_empty=True, pad=True)
    # query for 1 file
    q1, q2 = st.query_annotated(sel[0], sel[1], label=[2,4])
    d = '''filename sel_id start  end
f0.wav   0         0.0  5.1
f0.wav   1         4.0  9.1
f1.wav   0         0.0  5.1
f1.wav   1         4.0  9.1'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    pd.testing.assert_frame_equal(q1, ans[q1.columns.values])
    d = '''filename sel_id annot_id label  start  end  
f0.wav   0      1             2    3.0  5.1
f0.wav   1      1             2    0.0  2.3
f1.wav   0      0             4    1.0  4.3
f1.wav   0      1             2    4.0  5.1
f1.wav   1      0             4    0.0  0.3
f1.wav   1      1             2    0.0  3.3'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1,2])
    pd.testing.assert_frame_equal(q2, ans[q2.columns.values])

def test_file_duration_table(five_time_stamped_wave_files):
    """ Test that we can generate a file duration table""" 
    df = st.file_duration_table(five_time_stamped_wave_files)
    d = '''filename,duration
empty_HMS_12_ 5_ 0__DMY_23_ 2_84.wav,0.5
empty_HMS_12_ 5_ 1__DMY_23_ 2_84.wav,0.5
empty_HMS_12_ 5_ 2__DMY_23_ 2_84.wav,0.5
empty_HMS_12_ 5_ 3__DMY_23_ 2_84.wav,0.5
empty_HMS_12_ 5_ 4__DMY_23_ 2_84.wav,0.5'''
    ans = pd.read_csv(StringIO(d))
    pd.testing.assert_frame_equal(df, ans[df.columns.values])



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'audio.spectrogram' module within the ketos library
"""
import pytest
import numpy as np
import copy
import os
from ketos.audio.spectrogram import MagSpectrogram,\
    PowerSpectrogram, MelSpectrogram, Spectrogram, CQTSpectrogram
from ketos.audio.utils.axis import LinearAxis

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


def test_init_spec(spec_image_with_attrs):
    """Test that we can initialize an instance of the Spectrogram class"""
    img, dt, ax = spec_image_with_attrs
    spec = Spectrogram(data=img, time_res=dt, spec_type='MagSpectrogram', freq_ax=ax)
    assert np.all(spec.data == img)
    assert spec.type == 'MagSpectrogram'

def test_init_mag_spec():
    """Test that we can initialize an instance of the MagSpectrogram class
       from keyword arguments"""
    img = np.ones((20,10))
    spec = MagSpectrogram(data=img, time_res=1.0, freq_min=100, freq_res=4)
    assert np.all(spec.data == img)
    assert spec.type == 'MagSpectrogram'

def test_copy_spec(spec_image_with_attrs):
    """Test that we can make a copy of spectrogram"""
    img, dt, ax = spec_image_with_attrs
    spec = Spectrogram(data=img, time_res=dt, spec_type='MagSpectrogram', freq_ax=ax)
    spec2 = spec.deepcopy()
    assert np.all(spec.data == spec2.data)
    spec2.data += 1.5 #modify copied image
    spec2.time_ax.x_min += 30. #modify copied time axis
    assert np.all(spec.data + 1.5 == spec2.data) #check that original image was not affected
    assert spec.time_ax.min() + 30. == spec2.time_ax.min() #check that original time axis was not affected

def test_mag_spec_of_sine_wave(sine_audio):
    """Test that we can compute the magnitude spectrogram of a sine wave"""
    duration = sine_audio.duration()
    win = duration / 4
    step = duration / 10
    spec = MagSpectrogram.from_waveform(audio=sine_audio, window=win, step=step)
    assert spec.time_res() == step
    assert spec.freq_min() == 0    
    freq = np.argmax(spec.data, axis=1)
    freqHz = freq * spec.freq_res()
    assert np.all(np.abs(freqHz - 2000) < spec.freq_res())

def test_power_spec_of_sine_wave(sine_audio):
    """Test that we can compute the power spectrogram of a sine wave"""
    duration = sine_audio.duration()
    win = duration / 4
    step = duration / 10
    spec = PowerSpectrogram.from_waveform(audio=sine_audio, window=win, step=step)
    assert spec.time_res() == step
    assert spec.freq_min() == 0
    assert spec.type == 'PowerSpectrogram'
    freq = np.argmax(spec.data, axis=1)
    freqHz = freq * spec.freq_res()
    assert np.all(np.abs(freqHz - 2000) < spec.freq_res())

def test_mel_spec_of_sine_wave(sine_audio):
    """Test that we can compute the Mel spectrogram of a sine wave"""    
    duration = sine_audio.duration()
    win = duration / 4
    step = duration / 10
    spec = MelSpectrogram.from_waveform(audio=sine_audio, window=win, step=step)
    assert spec.time_res() == step
    assert spec.freq_min() == 0    

def test_cqt_spec_of_sine_wave(sine_audio):
    """Test that we can compute the CQT spectrogram of a sine wave"""    
    duration = sine_audio.duration()
    step = duration / 10
    spec = CQTSpectrogram.from_waveform(audio=sine_audio, step=step, bins_per_oct=64, freq_min=1, freq_max=4000)
    assert spec.freq_min() == 1
    freq = np.argmax(spec.data, axis=1)
    freqHz = spec.freq_ax.low_edge(freq)
    assert np.all(np.abs(freqHz - 2000) < 2 * spec.freq_ax.bin_width(freq))
    
def test_add_preserves_shape(sine_audio):
    """Test that when we add a spectrogram the shape of the present instance is preserved"""
    spec1 = MagSpectrogram.from_waveform(audio=sine_audio, window=0.2, step=0.05)
    orig_shape = spec1.data.shape
    spec2 = MagSpectrogram.from_waveform(audio=sine_audio, window=0.2, step=0.05)
    spec2.crop(start=1.0, end=2.5, freq_min=1000, freq_max=4000)
    spec1.add(spec2)
    assert spec1.data.shape == orig_shape

def test_add(sine_audio):
    """Test that when we add two spectrograms, we get the expected result"""
    spec1 = MagSpectrogram.from_waveform(audio=sine_audio, window=0.2, step=0.05)
    spec2 = MagSpectrogram.from_waveform(audio=sine_audio, window=0.2, step=0.05)
    spec12 = spec1.add(spec2, offset=1.0, scale=1.3, make_copy=True)
    bx = spec12.time_ax.bin(1.0)
    assert np.all(np.abs(spec12.data[:bx] - spec2.data[:bx]) < 0.001) # values before t=1.0 s are unchanged
    sum_spec = spec1.data[bx:] + 1.3 * spec2.data[:spec1.data.shape[0]-bx]
    assert np.all(np.abs(spec12.data[bx:] - sum_spec) < 0.001) # values outside addition region have changed

def test_cropped_mag_spec_has_correct_frequency_axis_range(sine_audio):
    """Test that when we crop a spectrogram along the frequency axis, we get the correct range"""
    spec = MagSpectrogram.from_waveform(audio=sine_audio, window=0.2, step=0.05)
    spec.crop(freq_max=4000)
    assert pytest.approx(spec.freq_max(), 4000, 2*spec.freq_res())
    spec.crop(freq_min=1000)
    assert pytest.approx(spec.freq_min(), 1000, 2*spec.freq_res())

def test_blur_time_axis():
    """Test that blurring along time axis gives expected results"""
    img = np.zeros((21,21))
    img[10,10] = 1
    ax = ax = LinearAxis(bins=img.shape[1], extent=(0., 21.), label='Frequency (Hz)')
    spec = Spectrogram(data=img, time_res=1, spec_type='MagSpectrogram', freq_ax=ax)
    sig = 2.0
    spec.blur(sigma_time=sig, sigma_freq=0.01)
    xy = spec.data / np.max(spec.data)
    x = xy[:,10]
    assert x[10] == pytest.approx(1, rel=0.001)
    assert x[9] == pytest.approx(np.exp(-pow(1,2)/(2.*pow(sig,2))), rel=0.001)
    assert x[8] == pytest.approx(np.exp(-pow(2,2)/(2.*pow(sig,2))), rel=0.001)    
    assert xy[10,9] == pytest.approx(0, abs=0.001) 

def test_blur_freq_axis():
    """Test that blurring along frequency axis gives expected results"""
    img = np.zeros((21,21))
    img[10,10] = 1
    ax = ax = LinearAxis(bins=img.shape[1], extent=(0., 21.), label='Frequency (Hz)')
    spec = Spectrogram(data=img, time_res=1, spec_type='MagSpectrogram', freq_ax=ax)
    sig = 4.2
    spec.blur(sigma_time=0.01, sigma_freq=sig)
    xy = spec.data / np.max(spec.data)
    y = xy[10,:]
    assert y[10] == pytest.approx(1, rel=0.001)
    assert y[9] == pytest.approx(np.exp(-pow(1,2)/(2.*pow(sig,2))), rel=0.001)
    assert y[8] == pytest.approx(np.exp(-pow(2,2)/(2.*pow(sig,2))), rel=0.001)    
    assert xy[9,10] == pytest.approx(0, abs=0.001) 

def test_recover_waveform(sine_audio):
    """Test that the recovered waveform has the correct sampling rate"""
    sine_audio.resample(new_rate=16000)
    duration = sine_audio.duration()
    win = duration / 4
    step = duration / 10
    spec = MagSpectrogram.from_waveform(audio=sine_audio, window=win, step=step)
    audio = spec.recover_waveform(num_iters=10)
    assert audio.rate == sine_audio.rate

def test_recover_waveform_after_time_crop(sine_audio):
    """Test that the recovered waveform from a time-cropped spectrogram has the correct sampling rate"""
    sine_audio.resample(new_rate=16000)
    win = 0.2
    step = 0.02
    spec = MagSpectrogram.from_waveform(audio=sine_audio, window=win, step=step)
    spec.crop(start=0.4, end=2.7)
    audio = spec.recover_waveform(num_iters=10)
    assert audio.rate == pytest.approx(sine_audio.rate, abs=0.1)

def test_recover_waveform_after_freq_crop(sine_audio):
    """Test that the recovered waveform from a frequency-cropped spectrogram has the correct sampling rate"""
    sine_audio.resample(new_rate=16000)
    win = 0.2
    step = 0.02
    spec = MagSpectrogram.from_waveform(audio=sine_audio, window=win, step=step)
    spec.crop(freq_min=200, freq_max=2300)
    audio = spec.recover_waveform(num_iters=10)
    assert audio.rate == pytest.approx(2*2300, abs=0.5)

def test_mag_from_wav(sine_wave_file):
    # duration is even integer multiply of step size
    spec = MagSpectrogram.from_wav(sine_wave_file, window=0.2, step=0.02)
    assert spec.time_res() == 0.02
    assert spec.duration() == 3.0
    # duration is not integer even multiply of step size, but adjust duration automatically
    spec = MagSpectrogram.from_wav(sine_wave_file, window=0.2, step=0.01)
    assert spec.time_res() == pytest.approx(0.01, abs=0.001)
    assert spec.duration() == pytest.approx(3.0, abs=0.01)
    # segment is empty raises assertion error
    with pytest.raises(AssertionError):
        spec = MagSpectrogram.from_wav(sine_wave_file, window=0.2, step=0.01, offset=4.0)
    # duration can be less than full length
    spec = MagSpectrogram.from_wav(sine_wave_file, window=0.2, step=0.02, duration=2.14)
    assert spec.time_res() == 0.02
    assert spec.duration() == 2.14
    # specify both offset and duration
    spec = MagSpectrogram.from_wav(sine_wave_file, window=0.2, step=0.02, offset=0.13, duration=2.14)
    assert spec.time_res() == 0.02
    assert spec.duration() == 2.14
    assert spec.offset == 0.13
    # check file name
    assert spec.filename == 'sine_wave.wav'

def test_mag_from_wav_id(sine_wave_file):
    """ Test that mag spectrogram created with from_wav method 
        has expected filename attribute"""
    # duration is even integer multiply of step size
    spec = MagSpectrogram.from_wav(sine_wave_file, window=0.2, step=0.02)
    assert spec.filename == 'sine_wave.wav'
    spec = MagSpectrogram.from_wav(sine_wave_file, window=0.2, step=0.02, id='test/audio.wav')
    assert spec.filename == 'test/audio.wav'

def test_cqt_from_wav(sine_wave_file):
    # zero offset
    spec = CQTSpectrogram.from_wav(sine_wave_file, step=0.01, freq_min=1, freq_max=300, bins_per_oct=32)
    assert spec.duration() == pytest.approx(3.0, abs=0.01)
    # non-zero offset
    offset = 1.0
    spec = CQTSpectrogram.from_wav(sine_wave_file, step=0.01, freq_min=1, freq_max=300, bins_per_oct=32, offset=1.0)
    assert spec.offset == offset
    assert spec.duration() == pytest.approx(3.0 - offset, abs=0.01)
    # duration is less than segment length
    duration = 1.1
    spec = CQTSpectrogram.from_wav(sine_wave_file, step=0.01, freq_min=1, freq_max=300, bins_per_oct=32, duration=duration)
    assert spec.duration() == pytest.approx(duration, abs=0.01)
    # step size is not divisor of duration
    spec = CQTSpectrogram.from_wav(sine_wave_file, step=0.017, freq_min=1, freq_max=300, bins_per_oct=32)
    assert spec.duration() == pytest.approx(3.0, abs=0.02)

def test_cqt_from_wav_id(sine_wave_file):
    """ Test that cqt spectrogram created with from_wav method 
        has expected filename attribute"""
    # duration is even integer multiply of step size
    spec = CQTSpectrogram.from_wav(sine_wave_file, step=0.01, freq_min=1, freq_max=300, bins_per_oct=32)
    assert spec.filename == 'sine_wave.wav'
    spec = CQTSpectrogram.from_wav(sine_wave_file, step=0.01, freq_min=1, freq_max=300, bins_per_oct=32, id='test/audio.wav')
    assert spec.filename == 'test/audio.wav'

def test_mel_from_wav(sine_wave_file):
    spec = MelSpectrogram.from_wav(sine_wave_file, window=0.2, step=0.02)
    assert spec.time_res() == 0.02

def test_stack_mag_specs():
    img = np.ones((20,10))
    s = MagSpectrogram(data=img, time_res=1.0, freq_min=100, freq_res=4, window_func='hamming')
    s.annotate(label=1, start=0.2, end=1.3)
    s.annotate(label=2, start=1.8, end=2.2)
    stacked = MagSpectrogram.stack([s, s])
    assert stacked.ndim == 2
    assert np.ndim(stacked.get_data()) == 3
    for i in range(2):
        assert np.all(np.abs(stacked.get_data(i) - img) < 1e-6)
        assert len(stacked.get_annotations(i)) == 2
        assert stacked.window_func == 'hamming'




# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Waveform module within the ketos library

    This module provides utilities to work with audio data.

    Contents:
        Waveform class
"""
import os
import numpy as np
import librosa
import scipy.io.wavfile as wave
from scipy import interpolate
import matplotlib.pyplot as plt
from ketos.utils import ensure_dir, morlet_func
from ketos.data_handling.data_handling import read_wave
from ketos.audio.annotation import AnnotationHandler
from ketos.audio.utils.axis import LinearAxis
from ketos.audio.base_audio import BaseAudio, segment_data
import ketos.audio.utils.misc as aum

class Waveform(BaseAudio):
    """ Audio signal

        Args:
            rate: float
                Sampling rate in Hz
            data: numpy array
                Audio data 
            filename: str
                Filename of the original audio file, if available (optional)
            offset: float
                Position within the original audio file, in seconds 
                measured from the start of the file. Defaults to 0 if not specified.
            label: int
                Spectrogram label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional

        Attributes:
            rate: float
                Sampling rate in Hz
            data: 1numpy array
                Audio data 
            time_ax: LinearAxis
                Axis object for the time dimension
            filename: str
                Filename of the original audio file, if available (optional)
            offset: float
                Position within the original audio file, in seconds 
                measured from the start of the file. Defaults to 0 if not specified.
            label: int
                Spectrogram label.
            annot: AnnotationHandler
                AnnotationHandler object.
    """
    def __init__(self, data, time_res=None, filename='', offset=0, label=None, annot=None, **kwargs):
        assert time_res is not None or 'rate' in kwargs, "either time_res or rate must be specified"

        if time_res is None:
            self.rate = kwargs['rate']
        else:
            self.rate = 1. / time_res

        super().__init__(data=data, time_res=1./self.rate, ndim=1, filename=filename, offset=offset, label=label, annot=annot)

    @classmethod
    def from_wav(cls, path, channel=0, rate=None, offset=0, duration=None, resample_method='scipy',
        id=None):
        """ Load audio data from wave file.

            Args:
                path: str
                    Path to input wave file
                channel: int
                    In the case of stereo recordings, this argument is used 
                    to specify which channel to read from. Default is 0.
                rate: float
                    Desired sampling rate in Hz. If None, the original sampling rate will be used
                offset: float
                    Position within the original audio file, in seconds 
                    measured from the start of the file. Defaults to 0 if not specified.
                duration: float
                    Length in seconds.
                resample_method: str
                    Resampling method. Only relevant if `rate` is specified. Options are
                        * kaiser_best
                        * kaiser_fast
                        * scipy (default)
                        * polyphase
                    See https://librosa.github.io/librosa/generated/librosa.core.resample.html 
                    for details on the individual methods.
                id: str
                    Unique identifier (optional). If None, the filename will be used.

            Returns:
                Instance of Waveform
                    Audio signal

            Example:
                >>> from ketos.audio.waveform import Waveform
                >>> # read audio signal from wav file
                >>> a = Waveform.from_wav('ketos/tests/assets/grunt1.wav')
                >>> # show signal
                >>> fig = a.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/audio_grunt1.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/audio_grunt1.png
        """
        rate_orig = librosa.get_samplerate(path)
        start = aum.num_samples(offset, rate_orig)
        if duration:
            stop = aum.num_samples(offset + duration, rate_orig)
        else:
            stop = None

        rate_orig, data = read_wave(file=path, channel=channel, start=start, stop=stop)

        if rate and rate != rate_orig:
            data = librosa.core.resample(data, orig_sr=rate_orig, target_sr=rate, res_type=resample_method)
        else:
            rate = rate_orig

        # pad with zeros to achieve desired duration, if necessary
        if duration is not None:
            num_sampl = int(duration * rate - data.shape[0])
            if num_sampl > 0:
                data = np.pad(data, pad_width=((0,num_sampl)), mode='constant')

        if id is None: id = os.path.basename(path)
            
        return cls(rate=rate, data=data, filename=id, offset=offset)

    @classmethod
    def gaussian_noise(cls, rate, sigma, samples, filename="gaussian_noise"):
        """ Generate Gaussian noise signal

            Args:
                rate: float
                    Sampling rate in Hz
                sigma: float
                    Standard deviation of the signal amplitude
                samples: int
                    Length of the audio signal given as the number of samples
                filename: str
                    Meta-data string (optional)

            Returns:
                Instance of Waveform
                    Audio signal sampling of Gaussian noise

            Example:
                >>> from ketos.audio.waveform import Waveform
                >>> # create gaussian noise with sampling rate of 10 Hz, standard deviation of 2.0 and 1000 samples
                >>> a = Waveform.gaussian_noise(rate=10, sigma=2.0, samples=1000)
                >>> # show signal
                >>> fig = a.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/audio_noise.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/audio_noise.png
        """        
        assert sigma > 0, "sigma must be strictly positive"

        y = np.random.normal(loc=0, scale=sigma, size=samples)
        return cls(rate=rate, data=y, filename=filename)

    @classmethod
    def morlet(cls, rate, frequency, width, samples=None, height=1, displacement=0, dfdt=0, filename="morlet"):
        """ Audio signal with the shape of the Morlet wavelet

            Uses :func:`util.morlet_func` to compute the Morlet wavelet.

            Args:
                rate: float
                    Sampling rate in Hz
                frequency: float
                    Frequency of the Morlet wavelet in Hz
                width: float
                    Width of the Morlet wavelet in seconds (sigma of the Gaussian envelope)
                samples: int
                    Length of the audio signal given as the number of samples (if no value is given, samples = 6 * width * rate)
                height: float
                    Peak value of the audio signal
                displacement: float
                    Peak position in seconds
                dfdt: float
                    Rate of change in frequency as a function of time in Hz per second.
                    If dfdt is non-zero, the frequency is computed as 

                        f = frequency + (time - displacement) * dfdt 

                filename: str
                    Meta-data string (optional)

            Returns:
                Instance of Waveform
                    Audio signal sampling of the Morlet wavelet 

            Examples:
                >>> from ketos.audio.waveform import Waveform
                >>> # create a Morlet wavelet with frequency of 3 Hz and 1-sigma width of envelope set to 2.0 seconds
                >>> wavelet1 = Waveform.morlet(rate=100., frequency=3., width=2.0)
                >>> # show signal
                >>> fig = wavelet1.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/morlet_standard.png")

                .. image:: ../../../../ketos/tests/assets/tmp/morlet_standard.png

                >>> # create another wavelet, but with frequency increasing linearly with time
                >>> wavelet2 = Waveform.morlet(rate=100., frequency=3., width=2.0, dfdt=0.3)
                >>> # show signal
                >>> fig = wavelet2.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/morlet_dfdt.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/morlet_dfdt.png
        """        
        if samples is None:
            samples = int(6 * width * rate)

        N = int(samples)

        # compute Morlet function at N equally spaced points
        dt = 1. / rate
        stop = (N-1.)/2. * dt
        start = -stop
        time = np.linspace(start, stop, N)
        y = morlet_func(time=time, frequency=frequency, width=width, displacement=displacement, norm=False, dfdt=dfdt)        
        y *= height
        
        return cls(rate=rate, data=np.array(y), filename=filename)

    @classmethod
    def cosine(cls, rate, frequency, duration=1, height=1, displacement=0, filename="cosine"):
        """ Audio signal with the shape of a cosine function

            Args:
                rate: float
                    Sampling rate in Hz
                frequency: float
                    Frequency of the Morlet wavelet in Hz
                duration: float
                    Duration of the signal in seconds
                height: float
                    Peak value of the audio signal
                displacement: float
                    Phase offset in fractions of 2*pi
                filename: str
                    Meta-data string (optional)

            Returns:
                Instance of Waveform
                    Audio signal sampling of the cosine function 

            Examples:
                >>> from ketos.audio.waveform import Waveform
                >>> # create a Cosine wave with frequency of 7 Hz
                >>> cos = Waveform.cosine(rate=1000., frequency=7.)
                >>> # show signal
                >>> fig = cos.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/cosine_audio.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/cosine_audio.png
        """        
        N = int(duration * rate)

        # compute cosine function at N equally spaced points
        dt = 1. / rate
        stop = (N-1.)/2. * dt
        start = -stop
        time = np.linspace(start, stop, N)
        x = (time * frequency + displacement) * 2 * np.pi
        y = height * np.cos(x)
        
        return cls(rate=rate, data=np.array(y), filename=filename)

    def get_attrs(self):
        return {'rate':self.rate, 'type':self.__class__.__name__}

    def get_data(self, id=0):
        """ Get the underlying data numpy array.

            Args:
                id: int
                    Audio signal ID. Only relevant if the Waveform object 
                    contains multiple, stacked audio signals.

            Returns:
                d: numpy array
                    Data
        """
        if id is None or np.ndim(self.data) == 1:
            d = self.data
        else:
            d = self.data[:,id]

        return d

    def segment(self, window, step=None):
        """ Divide the time axis into segments of uniform length, which may or may 
            not be overlapping.

            Window length and step size are converted to the nearest integer number 
            of time steps.

            If necessary, the audio signal will be padded with zeros at the end to 
            ensure that all segments have an equal number of samples. 

            Args:
                window: float
                    Length of each segment in seconds.
                step: float
                    Step size in seconds.

            Returns:
                segs: Waveform
                    Stacked audio signals

            Example:
                >>> from ketos.audio.waveform import Waveform
                >>> # create a morlet wavelet
                >>> mor = Waveform.morlet(rate=100, frequency=5, width=0.5)
                >>> mor.duration()
                3.0
                >>> # segment into 2-s wide frames, using a step size of 1 s
                >>> segs = mor.segment(window=2., step=1.)
                >>> # show the segments
                >>> fig0 = segs.plot(0)
                >>> fig0.savefig("ketos/tests/assets/tmp/morlet_segmented_0.png")
                >>> fig1 = segs.plot(1)
                >>> fig1.savefig("ketos/tests/assets/tmp/morlet_segmented_1.png")
                >>> plt.close(fig0)
                >>> plt.close(fig1)

                .. image:: ../../../../ketos/tests/assets/tmp/morlet_segmented_0.png

                .. image:: ../../../../ketos/tests/assets/tmp/morlet_segmented_1.png
        """              
        segs, filename, offset, label, annot = segment_data(self, window, step)

        d = self.__class__(data=segs, rate=self.rate, filename=filename,\
            offset=offset, label=label, annot=annot)

        return d

    def to_wav(self, path, auto_loudness=True):
        """ Save audio signal to wave file

            Args:
                path: str
                    Path to output wave file
                auto_loudness: bool
                    Automatically amplify the signal so that the 
                    maximum amplitude matches the full range of 
                    a 16-bit wav file (32760)
        """        
        ensure_dir(path)
        
        if auto_loudness:
            m = max(1, np.max(np.abs(self.data)))
            s = 32760 / m
        else:
            s = 1

        wave.write(filename=path, rate=int(self.rate), data=(s*self.data).astype(dtype=np.int16))

    def plot(self, id=0, show_annot=False):
        """ Plot the data with proper axes ranges and labels.

            Optionally, also display annotations as boxes superimposed on the data.

            Note: The resulting figure can be shown (fig.show())
            or saved (fig.savefig(file_name))

            Args:
                id: int
                    ID of data array to be plotted. Only relevant if the object 
                    contains multiple, stacked data arrays.
                show_annot: bool
                    Display annotations
            
            Returns:
                fig: matplotlib.figure.Figure
                    Figure object.

            Example:            
                >>> from ketos.audio.waveform import Waveform
                >>> # create a morlet wavelet
                >>> a = Waveform.morlet(rate=100, frequency=5, width=1)
                >>> # plot the wave form
                >>> fig = a.plot()
                >>> plt.close(fig)

                .. image:: ../../_static/morlet.png
        """
        fig, ax = super().plot(id=id)

        y = self.get_data(id)

        x = np.linspace(start=0, stop=self.duration(), num=self.data.shape[0])
        ax.plot(x, y)
        ax.set_ylabel('Amplitude')

        # superimpose annotation boxes
        if show_annot: self._draw_annot_boxes(ax,id)

        #fig.tight_layout()
        return fig

    def _draw_annot_boxes(self, ax, id=0):
        """Draws annotations boxes on top of the spectrogram

            Args:
                ax: matplotlib.axes.Axes
                    Axes object
                id: int
                    Object ID
        """
        annots = self.get_annotations(id=id)
        if annots is None: return
        y1, y2 = ax.get_ylim()
        for idx,annot in annots.iterrows():
            x1 = annot['start']
            x2 = annot['end']
            box = patches.Rectangle((x1,y1),x2-x1,y2-y1,linewidth=1,edgecolor='C1',facecolor='none')
            ax.add_patch(box)
            ax.text(x1, y2, int(annot['label']), ha='left', va='bottom', color='C1')

    def append(self, signal, n_smooth=0):
        """ Append another audio signal to the present instance.

            The two audio signals must have the same samling rate.
            
            If n_smooth > 0, a smooth transition is made between the 
            two signals in a overlap region of length n_smooth.

            Note that the current implementation of the smoothing procedure is 
            quite slow, so it is advisable to use small value for n_smooth.

            Args:
                signal: Waveform
                    Audio signal to be appended.
                n_smooth: int
                    Width of the smoothing/overlap region (number of samples).

            Returns:
                None

            Example:
                >>> from ketos.audio.waveform import Waveform
                >>> # create a morlet wavelet
                >>> mor = Waveform.morlet(rate=100, frequency=5, width=1)
                >>> # create a cosine wave
                >>> cos = Waveform.cosine(rate=100, frequency=3, duration=4)
                >>> # append the cosine wave to the morlet wavelet, using a overlap of 100 bins
                >>> mor.append(signal=cos, n_smooth=100)
                >>> # show the wave form
                >>> fig = mor.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/morlet_cosine.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/morlet_cosine.png
        """   
        assert self.rate == signal.rate, "Cannot merge audio signals with different sampling rates."

        # if appending signal to itself, make a copy
        if signal is self:
            signal = self.deepcopy()

        # ensure that overlap region is shorter than either signal
        n_smooth = min(n_smooth, len(self.data) - 1)
        n_smooth = min(n_smooth, len(signal.data) - 1)

        if n_smooth == 0:
            self.data = np.concatenate([self.data, signal.data], axis=0)

        else:# smoothly join
            a = self.data[:-n_smooth]
            ao = self.data[-n_smooth:]
            bo = signal.data[:n_smooth]
            b = signal.data[n_smooth:]

            # compute values in overlap region
            c = np.empty(n_smooth)
            for i in range(n_smooth):
                w = _smoothclamp(i, 0, n_smooth-1)
                c[i] = (1.-w) * ao[i] + w * bo[i]
            
            self.data = np.concatenate([a,c,b], axis=0)
        
        # re-init time axis
        length = self.data.shape[0] / self.rate
        self.time_ax = LinearAxis(bins=self.data.shape[0], extent=(0., length), label='Time (s)') 

    def add_gaussian_noise(self, sigma):
        """ Add Gaussian noise to the signal

            Args:
                sigma: float
                    Standard deviation of the gaussian noise

            Example:
                >>> from ketos.audio.waveform import Waveform
                >>> # create a morlet wavelet
                >>> morlet = Waveform.morlet(rate=100, frequency=2.5, width=1)
                >>> morlet_pure = morlet.deepcopy() # make a copy
                >>> # add some noise
                >>> morlet.add_gaussian_noise(sigma=0.3)
                >>> # show the wave form
                >>> fig = morlet_pure.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/morlet_wo_noise.png")
                >>> fig = morlet.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/morlet_w_noise.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/morlet_wo_noise.png

                .. image:: ../../../../ketos/tests/assets/tmp/morlet_w_noise.png
        """
        noise = Waveform.gaussian_noise(rate=self.rate, sigma=sigma, samples=len(self.data))
        self.add(noise)

    def add(self, signal, offset=0, scale=1):
        """ Add the amplitudes of the two audio signals.
        
            The audio signals must have the same sampling rates.
            The summed signal always has the same length as the present instance.
            If the audio signals have different lengths and/or a non-zero delay is selected, 
            only the overlap region will be affected by the operation.
            If the overlap region is empty, the original signal is unchanged.

            Args:
                signal: Waveform
                    Audio signal to be added
                offset: float
                    Shift the audio signal by this many seconds
                scale: float
                    Scaling factor applied to signal that is added

            Example:
                >>> from ketos.audio.waveform import Waveform
                >>> # create a cosine wave
                >>> cos = Waveform.cosine(rate=100, frequency=1., duration=4)
                >>> # create a morlet wavelet
                >>> mor = Waveform.morlet(rate=100, frequency=7., width=0.5)
                >>> mor.duration()
                3.0
                >>> # add the morlet wavelet on top of the cosine, with a shift of 1.5 sec and a scaling factor of 0.5
                >>> cos.add(signal=mor, offset=1.5, scale=0.5)
                >>> # show the wave form
                >>> fig = cos.plot()
                >>> fig.savefig("ketos/tests/assets/tmp/morlet_cosine_added.png")
                >>> plt.close(fig)

                .. image:: ../../../../ketos/tests/assets/tmp/morlet_cosine_added.png
        """
        assert self.rate == signal.rate, "Cannot add audio signals with different sampling rates."

        # if appending signal to itself, make a copy
        if signal is self:
            signal = self.deepcopy()

        # convert to bin numbers
        bin_offset = self.time_ax.bin(offset, truncate=True)
        bin_start = self.time_ax.bin(-offset, truncate=True)

        # crop signal that is being added
        length = self.data.shape[0] - bin_offset
        signal = signal.crop(start=-offset, length=length)

        # add the two signals
        b = bin_offset
        bins = signal.data.shape[0]
        self.data[b:b+bins] = self.data[b:b+bins] + scale * signal.data

    def resample(self, new_rate):
        """ Resample the acoustic signal with an arbitrary sampling rate.

        Note: Code adapted from Kahl et al. (2017)
              Paper: http://ceur-ws.org/Vol-1866/paper_143.pdf
              Code:  https://github.com/kahst/BirdCLEF2017/blob/master/birdCLEF_spec.py  

        Args:
            new_rate: int
                New sampling rate in Hz
        """
        if len(self.data) < 2:
            self.rate = new_rate

        else:                
            orig_rate = self.rate
            sig = self.data

            duration = sig.shape[0] / orig_rate

            time_old  = np.linspace(0, duration, sig.shape[0])
            time_new  = np.linspace(0, duration, int(sig.shape[0] * new_rate / orig_rate))

            interpolator = interpolate.interp1d(time_old, sig.T)
            new_audio = interpolator(time_new).T

            new_sig = np.round(new_audio).astype(sig.dtype)

            self.rate = new_rate
            self.data = new_sig
            self.time_ax = LinearAxis(bins=self.data.shape[0], extent=(0., self.data.shape[0] / self.rate), label='Time (s)') #new time axis


def _smoothclamp(x, mi, mx): 
        """ Smoothing function
        """    
        return (lambda t: np.where(t < 0 , 0, np.where( t <= 1 , 3*t**2-2*t**3, 1 ) ) )( (x-mi)/(mx-mi) )



import pytest
import numpy as np
import tensorflow as tf
from ketos.neural_networks.dev_utils.nn_interface import RecipeCompat
from ketos.neural_networks.cnn import CNNArch, CNNInterface
from ketos.neural_networks.dev_utils.losses import FScoreLoss
import os
import tables
import json


current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


@pytest.fixture
def recipe_simple_dict():
    recipe = {'conv_set':[[64, False], [128, True], [256, True]],
               'dense_set': [512, 256],
               'n_classes':2,
               'optimizer': {'recipe_name':'Adam', 'parameters': {'learning_rate':0.005}},
               'loss_function': {'recipe_name':'FScoreLoss', 'parameters':{}},  
               'metrics': [{'recipe_name':'CategoricalAccuracy', 'parameters':{}}]

    }

    return recipe


@pytest.fixture
def recipe_simple():
    recipe = {'conv_set':[[64, False], [128, True], [256, True]],
               'dense_set': [512, 256],
               'n_classes':2,        
               'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
               'loss_function': RecipeCompat('FScoreLoss', FScoreLoss),  
               'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)]
        
    }
    return recipe


@pytest.fixture
def recipe_detailed_dict():
    recipe = {'convolutional_layers':  [{'n_filters':64, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True},
                                    {'n_filters':128, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True},
                                    {'n_filters':256, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True}],
              'dense_layers':[{'n_hidden':512, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    {'n_hidden':256, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    ],
               'n_classes':2,
               'optimizer': {'recipe_name':'Adam', 'parameters': {'learning_rate':0.005}},
               'loss_function': {'recipe_name':'FScoreLoss', 'parameters':{}},  
               'metrics': [{'recipe_name':'CategoricalAccuracy', 'parameters':{}}]

    }

    return recipe

@pytest.fixture
def recipe_detailed():
    recipe = {'convolutional_layers':  [{'n_filters':64, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True},
                                    {'n_filters':128, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True},
                                    {'n_filters':256, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True}],
              'dense_layers':[{'n_hidden':512, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    {'n_hidden':256, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    ],
               'n_classes':2,
               'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
               'loss_function': RecipeCompat('FScoreLoss', FScoreLoss),  
               'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)]
        
    }

    return recipe


def test_CNNArch():
    conv_layers =  [{'n_filters':64, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':None, 'batch_normalization':True},
                                    {'n_filters':128, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True},
                                    {'n_filters':256, "filter_shape":[3,3], 'strides':1, 'padding':'valid', 'activation':'relu', 'max_pool':{'pool_size':[2,2] , 'strides':[2,2]}, 'batch_normalization':True}]

    dense_layers = [{'n_hidden':512, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    {'n_hidden':256, 'activation':'relu', 'batch_normalization':True, 'dropout':0.5},
                                    ]
    
    cnn = CNNArch(convolutional_layers=conv_layers, dense_layers=dense_layers, n_classes=2)
    assert len(cnn.layers) == 2

    #convolutional block
    assert len(cnn.layers[0].layers) == 8
    assert isinstance(cnn.layers[0].layers[0], tf.keras.layers.Conv2D)
    assert isinstance(cnn.layers[0].layers[1], tf.keras.layers.BatchNormalization)
    assert isinstance(cnn.layers[0].layers[2], tf.keras.layers.Conv2D)
    assert isinstance(cnn.layers[0].layers[3], tf.keras.layers.MaxPooling2D)
    assert isinstance(cnn.layers[0].layers[4], tf.keras.layers.BatchNormalization)
    assert isinstance(cnn.layers[0].layers[5], tf.keras.layers.Conv2D)
    assert isinstance(cnn.layers[0].layers[6], tf.keras.layers.MaxPooling2D)
    assert isinstance(cnn.layers[0].layers[7], tf.keras.layers.BatchNormalization)

    #Dense block
    assert len(cnn.layers[1].layers) == 8
    assert isinstance(cnn.layers[1].layers[0], tf.keras.layers.Dense)
    assert isinstance(cnn.layers[1].layers[1], tf.keras.layers.BatchNormalization)
    assert isinstance(cnn.layers[1].layers[2], tf.keras.layers.Dropout)
    assert isinstance(cnn.layers[1].layers[3], tf.keras.layers.Dense)
    assert isinstance(cnn.layers[1].layers[4], tf.keras.layers.BatchNormalization)
    assert isinstance(cnn.layers[1].layers[5], tf.keras.layers.Dropout)
    assert isinstance(cnn.layers[1].layers[6], tf.keras.layers.Dense)
    assert isinstance(cnn.layers[1].layers[7], tf.keras.layers.Softmax)
   

def test_convolutional_layers_from_conv_set(recipe_simple, recipe_detailed):
    detailed_layers = CNNInterface._convolutional_layers_from_conv_set(recipe_simple['conv_set'])
    assert detailed_layers == recipe_detailed['convolutional_layers']
    

def test_dense_layers_from_dense_set(recipe_simple, recipe_detailed):
    detailed_layers = CNNInterface._dense_layers_from_dense_set(recipe_simple['dense_set'])
    assert detailed_layers == recipe_detailed['dense_layers']


 
def test_CNNInterface_build_from_recipe_simple(recipe_simple, recipe_detailed):
    cnn = CNNInterface._build_from_recipe(recipe_simple)

    assert cnn.optimizer.recipe_name == recipe_simple['optimizer'].recipe_name
    assert cnn.optimizer.instance.__class__ == recipe_simple['optimizer'].instance.__class__
    assert cnn.optimizer.args == recipe_simple['optimizer'].args

    assert cnn.loss_function.recipe_name == recipe_simple['loss_function'].recipe_name
    assert cnn.loss_function.instance.__class__ == recipe_simple['loss_function'].instance.__class__
    assert cnn.loss_function.args == recipe_simple['loss_function'].args

    assert cnn.metrics[0].recipe_name == recipe_simple['metrics'][0].recipe_name
    assert cnn.metrics[0].instance.__class__ == recipe_simple['metrics'][0].instance.__class__
    assert cnn.metrics[0].args == recipe_simple['metrics'][0].args

    assert cnn.conv_set == recipe_simple['conv_set']
    assert cnn.dense_set == recipe_simple['dense_set']
    assert cnn.n_classes ==  recipe_simple['n_classes']


def test_CNNInterface_build_from_recipe_simple_dict(recipe_simple_dict, recipe_simple, recipe_detailed):
    cnn = CNNInterface._build_from_recipe(recipe_simple_dict, recipe_compat=False)

    assert cnn.optimizer.recipe_name == recipe_simple['optimizer'].recipe_name
    assert cnn.optimizer.instance.__class__ == recipe_simple['optimizer'].instance.__class__
    assert cnn.optimizer.args == recipe_simple['optimizer'].args

    assert cnn.loss_function.recipe_name == recipe_simple['loss_function'].recipe_name
    assert cnn.loss_function.instance.__class__ == recipe_simple['loss_function'].instance.__class__
    assert cnn.loss_function.args == recipe_simple['loss_function'].args

    assert cnn.metrics[0].recipe_name == recipe_simple['metrics'][0].recipe_name
    assert cnn.metrics[0].instance.__class__ == recipe_simple['metrics'][0].instance.__class__
    assert cnn.metrics[0].args == recipe_simple['metrics'][0].args

    assert cnn.conv_set == recipe_simple['conv_set']
    assert cnn.dense_set == recipe_simple['dense_set']
    assert cnn.n_classes ==  recipe_simple['n_classes']


def test_CNNInterface_build_from_recipe_detailed(recipe_detailed):
    cnn = CNNInterface._build_from_recipe(recipe_detailed)

    assert cnn.optimizer.recipe_name == recipe_detailed['optimizer'].recipe_name
    assert cnn.optimizer.instance.__class__ == recipe_detailed['optimizer'].instance.__class__
    assert cnn.optimizer.args == recipe_detailed['optimizer'].args

    assert cnn.loss_function.recipe_name == recipe_detailed['loss_function'].recipe_name
    assert cnn.loss_function.instance.__class__ == recipe_detailed['loss_function'].instance.__class__
    assert cnn.loss_function.args == recipe_detailed['loss_function'].args

    assert cnn.metrics[0].recipe_name == recipe_detailed['metrics'][0].recipe_name
    assert cnn.metrics[0].instance.__class__ == recipe_detailed['metrics'][0].instance.__class__
    assert cnn.metrics[0].args == recipe_detailed['metrics'][0].args

    assert cnn.convolutional_layers == recipe_detailed['convolutional_layers']
    assert cnn.dense_layers == recipe_detailed['dense_layers']
    assert cnn.n_classes ==  recipe_detailed['n_classes']

def test_CNNInterface_build_from_recipe_detailed_dict(recipe_detailed, recipe_detailed_dict):
    cnn = CNNInterface._build_from_recipe(recipe_detailed_dict, recipe_compat=False)

    assert cnn.optimizer.recipe_name == recipe_detailed['optimizer'].recipe_name
    assert cnn.optimizer.instance.__class__ == recipe_detailed['optimizer'].instance.__class__
    assert cnn.optimizer.args == recipe_detailed['optimizer'].args

    assert cnn.loss_function.recipe_name == recipe_detailed['loss_function'].recipe_name
    assert cnn.loss_function.instance.__class__ == recipe_detailed['loss_function'].instance.__class__
    assert cnn.loss_function.args == recipe_detailed['loss_function'].args

    assert cnn.metrics[0].recipe_name == recipe_detailed['metrics'][0].recipe_name
    assert cnn.metrics[0].instance.__class__ == recipe_detailed['metrics'][0].instance.__class__
    assert cnn.metrics[0].args == recipe_detailed['metrics'][0].args

    assert cnn.convolutional_layers == recipe_detailed['convolutional_layers']
    assert cnn.dense_layers == recipe_detailed['dense_layers']
    assert cnn.n_classes ==  recipe_detailed['n_classes']

def test_write_recipe_simple(recipe_simple, recipe_simple_dict, recipe_detailed):
    cnn = CNNInterface._build_from_recipe(recipe_simple)
    written_recipe = cnn._extract_recipe_dict()

    #Even when the model is built from a simplified recipe, the detailed form will be included when writing the recipe again

    recipe_simple_dict['convolutional_layers'] = recipe_detailed['convolutional_layers']
    recipe_simple_dict['dense_layers'] = recipe_detailed['dense_layers']

    assert written_recipe == recipe_simple_dict
    

def test_read_recipe_simple_file(recipe_simple, recipe_simple_dict, recipe_detailed):
    path_to_recipe_file = os.path.join(path_to_tmp, "test_cnn_recipe_simple.json")
    cnn = CNNInterface._build_from_recipe(recipe_simple)
    #written_recipe = resnet.write_recipe()
    cnn.save_recipe_file(path_to_recipe_file)

    #Read recipe as a recipe dict
    read_recipe = cnn._read_recipe_file(path_to_recipe_file,return_recipe_compat=False)
    recipe_simple_dict['convolutional_layers'] = recipe_detailed['convolutional_layers']
    recipe_simple_dict['dense_layers'] = recipe_detailed['dense_layers']
    assert read_recipe == recipe_simple_dict
    

    #Read recipe as a recipe dict with RecipeCompat objects
    recipe_simple['convolutional_layers'] = recipe_detailed['convolutional_layers']
    recipe_simple['dense_layers'] = recipe_detailed['dense_layers']

    read_recipe = cnn._read_recipe_file(path_to_recipe_file,return_recipe_compat=True)
    assert read_recipe['optimizer'].recipe_name == recipe_simple['optimizer'].recipe_name
    assert read_recipe['optimizer'].instance.__class__ == recipe_simple['optimizer'].instance.__class__
    assert read_recipe['optimizer'].args == recipe_simple['optimizer'].args

    assert read_recipe['loss_function'].recipe_name == recipe_simple['loss_function'].recipe_name
    assert read_recipe['loss_function'].instance.__class__ == recipe_simple['loss_function'].instance.__class__
    assert read_recipe['loss_function'].args == recipe_simple['loss_function'].args
    
    assert read_recipe['metrics'][0].recipe_name == recipe_simple['metrics'][0].recipe_name
    assert read_recipe['metrics'][0].instance.__class__ == recipe_simple['metrics'][0].instance.__class__
    assert read_recipe['metrics'][0].args == recipe_simple['metrics'][0].args

    assert read_recipe['conv_set'] == recipe_simple['conv_set']
    assert read_recipe['dense_set'] == recipe_simple['dense_set']
    assert read_recipe['convolutional_layers'] == recipe_simple['convolutional_layers']
    assert read_recipe['dense_layers'] == recipe_simple['dense_layers']


def test_read_recipe_simple_detailed(recipe_detailed, recipe_detailed_dict):
    path_to_recipe_file = os.path.join(path_to_tmp, "test_cnn_recipe_detailed.json")
    cnn = CNNInterface._build_from_recipe(recipe_detailed)
    #written_recipe = resnet.write_recipe()
    cnn.save_recipe_file(path_to_recipe_file)

    #Read recipe as a recipe dict
    read_recipe = cnn._read_recipe_file(path_to_recipe_file,return_recipe_compat=False)
    recipe_detailed_dict['conv_set'] = None
    recipe_detailed_dict['dense_set'] = None
    assert read_recipe == recipe_detailed_dict
    
    #Read recipe as a recipe dict with RecipeCompat objects
    recipe_detailed['conv_set'] = None
    recipe_detailed['dense_set'] = None

    read_recipe = cnn._read_recipe_file(path_to_recipe_file,return_recipe_compat=True)
    assert read_recipe['optimizer'].recipe_name == recipe_detailed['optimizer'].recipe_name
    assert read_recipe['optimizer'].instance.__class__ == recipe_detailed['optimizer'].instance.__class__
    assert read_recipe['optimizer'].args == recipe_detailed['optimizer'].args

    assert read_recipe['loss_function'].recipe_name == recipe_detailed['loss_function'].recipe_name
    assert read_recipe['loss_function'].instance.__class__ == recipe_detailed['loss_function'].instance.__class__
    assert read_recipe['loss_function'].args == recipe_detailed['loss_function'].args
    
    assert read_recipe['metrics'][0].recipe_name == recipe_detailed['metrics'][0].recipe_name
    assert read_recipe['metrics'][0].instance.__class__ == recipe_detailed['metrics'][0].instance.__class__
    assert read_recipe['metrics'][0].args == recipe_detailed['metrics'][0].args

    assert read_recipe['conv_set'] == recipe_detailed['conv_set']
    assert read_recipe['dense_set'] == recipe_detailed['dense_set']
    assert read_recipe['convolutional_layers'] == recipe_detailed['convolutional_layers']
    assert read_recipe['dense_layers'] == recipe_detailed['dense_layers']


